# 和processing 模块有点重叠
from datetime import timedelta, datetime
import scipy.optimize as optimize
import pandas as pd
import numpy as np
import time


class Map(dict):
    """
    Example:
    m = Map({'first_name': 'Eduardo'}, last_name='Pool', age=24, sports=['Soccer'])
    """

    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v
                    if isinstance(v, dict):
                        self[k] = Map(v)

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v
                if isinstance(v, dict):
                    self[k] = Map(v)

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Map, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Map, self).__delitem__(key)
        del self.__dict__[key]


# 打印字典
def dprint(d):
    for key in d:
        print("      -  " + str(key) + ':' + str(d[key]))


def exeTime(func):
    def newFunc(*args, **args2):
        t0 = time.time()
        print("@%s, {%s} start" % (time.strftime("%X", time.localtime()), func.__name__))
        back = func(*args, **args2)
        print("@%s, {%s} end" % (time.strftime("%X", time.localtime()), func.__name__))
        print("@%.3fs taken for {%s}" % (time.time() - t0, func.__name__))
        return back
    
    return newFunc


def nearest(items, pivot):
    """
    the reusult will be a little smaller than pivot with the positive timedelta
    :param items: the samples
    :param pivot: the target
    :return:
    """
    return min(items, key=lambda x: abs((x + timedelta(1) - pivot)))


def get_end_date():
    """确定开始时间，计算最新的已经结算的交易日"""
    try:
        from pyfi import WindHelper
        # 确定结束时间
        # 结束时间为该合约的最后交易日和当前日期的最小值
        last_trade_date = WindHelper.t_days_offset(offset=0, cur_date=datetime.now())
        # 确定结束时间
        if datetime.now().hour >= 19:  # 以晚上19点为界限
            end_date = last_trade_date
        elif datetime.now().date() > last_trade_date.date():  # 当天不是交易日
            end_date = last_trade_date
        else:  # 既非节假日，且当然的数据也没有生成
            end_date = WindHelper.t_days_offset(offset=-1, cur_date=datetime.now())  # datetime类型
    except Exception as e:
        end_date = datetime.now()
    return end_date


def get_date_list(begin_date, end_date):
    # begin_date, end_date是形如‘20160601’的字符串或datetime格式
    date_l = [x for x in list(pd.date_range(start=begin_date, end=end_date))]
    return date_l


# 常用债券计算函数
def get_ytm_by_net(net, maturity, cupn, freq, par=100, guess=3):
    """
    根据净价, 待偿期限，票息，付息次数，面值计算债券的到期收益率
    :param net:
    :param par:
    :param maturity:
    :param cupn:
    :param freq:
    :param guess:
    :return:
    """
    yearsToNxCupn = maturity - 1. / freq * np.floor(freq * maturity)
    cupnCount = int(np.ceil(freq * maturity))
    coupon = cupn / 100. * par / freq
    if yearsToNxCupn == 0.:
        dt = [(i + 1.) / freq for i in range(cupnCount)]
    else:
        dt = [yearsToNxCupn + i / freq for i in range(cupnCount)]
    ytm_func = lambda y: sum([coupon / (1. + y / 100. / freq) ** (freq * t) for t in dt]) + \
                         par / (1. + y / 100. / freq) ** (freq * maturity) - (
                                 1. / freq - yearsToNxCupn) * cupn - net
    return optimize.newton(ytm_func, guess)


def get_dirty(ytm, maturity, cupn, freq, par=100.):
    """
    based on ytm, maturity, coupon and frequency to calculate the dirty price
    :param ytm:
    :param par:
    :param maturity:
    :param cupn: 3.65 instead of 0.0365
    :param freq:
    :return:
    """
    yearsToNxCupn = maturity - 1.0 / freq * np.floor(freq * maturity)
    cupnCount = int(np.floor(freq * maturity)) + 1
    coupon = cupn / 100. * par / freq
    if yearsToNxCupn == 0:
        dt = [(i + 1.) / freq for i in range(cupnCount)]  # 记录付息时间点
    else:
        dt = [yearsToNxCupn + i / freq for i in range(cupnCount)]  # 记录付息时间点
    return sum([coupon / (1. + ytm / 100. / freq) ** (freq * t) for t in dt]) + \
           par / (1. + ytm / 100. / freq) ** (freq * maturity)


def get_net(ytm, par, maturity, cupn, freq):
    daysToNxCupn = maturity - 1.0 / freq * np.floor(freq * maturity)
    dirty = get_dirty(ytm, maturity, cupn, freq, par)
    return dirty - (1. / freq - daysToNxCupn) * cupn


def get_years_to_nxcupn(maturity, freq):
    """
    calculate the period to next coupon marked by year unit
    :param maturity:
    :param freq:
    :return:
    """
    yearsToNxCupn = maturity - 1.0 / freq * np.floor(freq * maturity)
    return yearsToNxCupn


# 数据序列预处理模块,存放一些时间序列预处理的通用模块，一般会被data_agent调用
import datetime as dt
import pandas as pd
import numpy as np


def yoy2fixedbaseindex(orgfbi, yoylist, option1='index', option2='yoy', fbiloc='start'):
    """
    将同比序列转换为定基指数
    author: HuPeiran
    # input: orgfib, yoylist = pandas.data dataframe
    # option1 = 'index' OR 'mom', option2 = 'yoy' OR 'cumyoy', fbiloc = 'start' OR ?

    # option1 代表初始定基指数的形式，指数形式 - 'index'，环比形式 - 'mom'
    # option2 代表计算定基指数的数据形式，同比数据 - 'yoy'，累计同比数据 - 'cumyoy'
    # option3 用于指定初始定基指数和后面环比数据位置

    # 初始定基指数不能与同比数据叠加
    # 举例：初始定基指数'2011-01:2011-12'，同比数据'2012-01:2018-05'
    # 函数最后将初始定基指数和同比数据一并返回
    :param orgfbi:
    :param yoylist:
    :param option1:
    :param option2:
    :param fbiloc:
    :return:
    """
    startmonth = orgfbi.index[0].month
    endmonth = orgfbi.index[-1].month
    cyc = endmonth - startmonth + 1

    if option1 == 'mom':
        temp = [1 + i for i in orgfbi.iloc[:, 0].values]
        temp[0] = 1
        orgfbi_list = np.cumprod(temp)
        orgfbi_list = [i * 100 for i in orgfbi_list]
    elif option1 == 'index':
        start = orgfbi.iloc[:, 0].values[0]
        temp = [i / start * 100 for i in orgfbi.iloc[:, 0].values]
        orgfbi_list = temp

    startyear = yoylist.iloc[:, 0].index[0].year
    alldatelist = yoylist.index
    allyearlist = [i.year for i in alldatelist]
    temp = list()
    [temp.append(i) for i in allyearlist if not i in temp]  # 取yoy数据的所有年份
    allyearlist = temp

    len_ = len(yoylist[str(allyearlist[-1])])
    if len_ != cyc:
        newdate = pd.date_range(start=dt.date(allyearlist[0], startmonth, 1),
                                end=dt.date(allyearlist[-1], endmonth, 31), freq='M')
        adddata = [0] * (cyc * (len(allyearlist) - 1) + len_) + [np.nan] * (cyc - len_)
        adddata_dataf = pd.DataFrame(data=adddata, index=newdate, columns=['temp'])
        yoylist.columns = ['temp']
        yoylist_new = yoylist + adddata_dataf

    yoy_1year = list()
    yoy_1year.append(orgfbi_list)

    for year_count in allyearlist:
        yoy_1year.append(
            [float(i) / 100 + 1 for i in yoylist_new.loc[str(year_count), :].values])  # dataframe.values = array

    yoy_1year_array = np.array([i for i in yoy_1year])
    re = list(np.cumprod(yoy_1year_array.T, axis=1).T.reshape((1, cyc * (len(allyearlist) + 1)))[0])
    re[-(cyc - len_):] = []
    re = pd.DataFrame(data=re, index=list(orgfbi.index) + list(yoylist.index), columns=['FixedBaseIndex'])
    return re


def macro_adjust(yoy, cyoy, method=1):
    """
    author: Wang Luzhou
    update date: 2018-08-01
    introduction: 去除1月，2月信息用累计同比
    :param yoy:
    :param cyoy:
    :param method: 1:remove Jan data; 2: copy Dec data to Jan
    :return:

    """
    for i in range(len(yoy)):
        if yoy.index[i].month == 2:  # 2月份
            yoy[i] = cyoy[i]
        if yoy.index[i].month == 1:  # 1月份
            if method == 1 or i == 0:
                yoy.iloc[i] = np.nan
            elif method == 2:
                yoy.iloc[i] = yoy.iloc[i - 1]
    return yoy.dropna()


def ffill_mon_data(data):
    """
    author: Wang Luzhou
    update: 2018-08-14
    将数据进行月度级别的填充，并向前复制覆盖na值
    日度数据也可以转换为月度数据
    :param data:
    :return:

    Examples
    --------
    当我们调用人民币定存利率的时候，我们收到的是不规则的低频数据，
    因此需要依靠这个函数将数据转换为月度数据
    """
    return data.resample("M").last().ffill().dropna()


def get_supseason_ind(data):
    """
    输入的为环比数据,返回的是超季节性指数
     # input: data = pandas.series
    :return:
    """
    from pyfi import spring_month
    # 环比数据
    data = data
    # cpi_r = w.edb(codes=[code], begin_date=datetime(1995, 1, 1), end_date=cur_date).iloc[:, 0]
    # 标记春节， 该月有春节则标记True,否则标记False
    spring = pd.Series(index=data.index)
    for i in range(len(data)):
        if data.index[i].month == 1 or i == 0:
            spring_m = spring_month(data.index[i].year)
        if data.index[i].month == spring_m:
            spring[i] = True
        else:
            spring[i] = False
    # 标记完成
    supseason_ind = pd.Series(index=data.index)
    supseason_cum_ind = pd.Series(index=data.index)
    cpi_cum_r = pd.Series(index=data.index)
    window = 5  # 和过去五年作比较
    for i in range(12, len(data)):
        if cpi_cum_r.index[i].month == 1:
            cpi_cum_r[i] = data[i] / 100.0
        else:
            cpi_cum_r[i] = (1 + (0 if np.isnan(cpi_cum_r[i - 1]) else cpi_cum_r[i - 1])) * (1 + data[i] / 100.0) - 1
        past = data[:i]
        past_cum = cpi_cum_r[:i]
        past_spring = spring[:i]
        spring_cur = spring[i]
        chosen = past.loc[(past.index.month == data.index[i].month) & (past_spring == spring_cur)][-window:]
        chosen_cum = past_cum.loc[(past_cum.index.month == cpi_cum_r.index[i].month)][-window:].dropna()
        if len(chosen) > 0:
            avg = chosen.mean()
            supseason_ind[i] = data[i] - avg
        if len(chosen_cum) > 0:
            avg_cum = chosen_cum.mean()
            supseason_cum_ind[i] = cpi_cum_r[i] - avg_cum
    return supseason_ind, supseason_cum_ind



import datetime as dt
import pandas as pd
import numpy as np

"""
本模块主要是常见的宏观序列的处理函数
当前的主要功能是：
1. 获取定基指数
2. 1,2月数据进行处理
3. 将累计值转换为单月值
"""


def yoy2fixedbaseindex(orgfbi, yoylist, option1='index', option2='yoy', fbiloc='start'):
    """
    将同比序列转换为定基指数
    author: HuPeiran
    # input: orgfib, yoylist = pandas.data dataframe
    # option1 = 'index' OR 'mom', option2 = 'yoy' OR 'cumyoy', fbiloc = 'start' OR ?

    # option1 代表初始定基指数的形式，指数形式 - 'index'，环比形式 - 'mom'
    # option2 代表计算定基指数的数据形式，同比数据 - 'yoy'，累计同比数据 - 'cumyoy'
    # option3 用于指定初始定基指数和后面环比数据位置

    # 初始定基指数不能与同比数据叠加
    # 举例：初始定基指数'2011-01:2011-12'，同比数据'2012-01:2018-05'
    # 函数最后将初始定基指数和同比数据一并返回
    :param orgfbi:
    :param yoylist:
    :param option1:
    :param option2:
    :param fbiloc:
    :return:
    """
    startmonth = orgfbi.index[0].month
    endmonth = orgfbi.index[-1].month
    cyc = endmonth - startmonth + 1

    if option1 == 'mom':
        temp = [1 + i for i in orgfbi.iloc[:, 0].values]
        temp[0] = 1
        orgfbi_list = np.cumprod(temp)
        orgfbi_list = [i * 100 for i in orgfbi_list]
    elif option1 == 'index':
        start = orgfbi.iloc[:, 0].values[0]
        temp = [i / start * 100 for i in orgfbi.iloc[:, 0].values]
        orgfbi_list = temp

    startyear = yoylist.iloc[:, 0].index[0].year
    alldatelist = yoylist.index
    allyearlist = [i.year for i in alldatelist]
    temp = list()
    [temp.append(i) for i in allyearlist if not i in temp]  # 取yoy数据的所有年份
    allyearlist = temp

    len_ = len(yoylist[str(allyearlist[-1])])
    if len_ != cyc:
        newdate = pd.date_range(start=dt.date(allyearlist[0], startmonth, 1),
                                end=dt.date(allyearlist[-1], endmonth, 31), freq='M')
        adddata = [0] * (cyc * (len(allyearlist) - 1) + len_) + [np.nan] * (cyc - len_)
        adddata_dataf = pd.DataFrame(data=adddata, index=newdate, columns=['temp'])
        yoylist.columns = ['temp']
        yoylist_new = yoylist + adddata_dataf

    yoy_1year = list()
    yoy_1year.append(orgfbi_list)

    for year_count in allyearlist:
        yoy_1year.append(
            [float(i) / 100 + 1 for i in yoylist_new.loc[str(year_count), :].values])  # dataframe.values = array

    yoy_1year_array = np.array([i for i in yoy_1year])
    re = list(np.cumprod(yoy_1year_array.T, axis=1).T.reshape((1, cyc * (len(allyearlist) + 1)))[0])
    re[-(cyc - len_):] = []
    re = pd.DataFrame(data=re, index=list(orgfbi.index) + list(yoylist.index), columns=['FixedBaseIndex'])
    return re


def macro_adjust(yoy, cyoy, method=1):
    """
    author: Wang Luzhou
    update date: 2018-08-01
    introduction: 去除1月，2月信息用累计同比
    :param yoy:
    :param cyoy:
    :param method: 1:remove Jan data; 2: copy Dec data to Jan
    :return:

    """
    for i in range(len(yoy)):
        if yoy.index[i].month == 2:  # 2月份
            yoy[i] = cyoy[i]
        if yoy.index[i].month == 1:  # 1月份
            if method == 1 or i == 0:
                yoy.iloc[i] = np.nan
            elif method == 2:
                yoy.iloc[i] = yoy.iloc[i - 1]
    return yoy.dropna()


def ffill_mon_data(data):
    """
    author: Wang Luzhou
    update: 2018-08-14
    将数据进行月度级别的填充，并向前复制覆盖na值
    日度数据也可以转换为月度数据
    :param data:
    :return:

    Examples
    --------
    当我们调用人民币定存利率的时候，我们收到的是不规则的低频数据，
    因此需要依靠这个函数将数据转换为月度数据
    """
    return data.resample("M").last().ffill().dropna()


def get_supseason_ind(data):
    """
    输入的为环比数据,返回的是超季节性指数
     # input: data = pandas.series
    :return:
    """
    from pyfi import spring_month
    # 环比数据
    data = data
    # cpi_r = w.edb(codes=[code], begin_date=datetime(1995, 1, 1), end_date=cur_date).iloc[:, 0]
    # 标记春节， 该月有春节则标记True,否则标记False
    spring = pd.Series(index=data.index)
    for i in range(len(data)):
        if data.index[i].month == 1 or i == 0:
            spring_m = spring_month(data.index[i].year)
        if data.index[i].month == spring_m:
            spring[i] = True
        else:
            spring[i] = False
    # 标记完成
    supseason_ind = pd.Series(index=data.index)
    supseason_cum_ind = pd.Series(index=data.index)
    cpi_cum_r = pd.Series(index=data.index)
    window = 5  # 和过去五年作比较
    for i in range(12, len(data)):
        if cpi_cum_r.index[i].month == 1:
            cpi_cum_r[i] = data[i] / 100.0
        else:
            cpi_cum_r[i] = (1 + (0 if np.isnan(cpi_cum_r[i - 1]) else cpi_cum_r[i - 1])) * (1 + data[i] / 100.0) - 1
        past = data[:i]
        past_cum = cpi_cum_r[:i]
        past_spring = spring[:i]
        spring_cur = spring[i]
        chosen = past.loc[(past.index.month == data.index[i].month) & (past_spring == spring_cur)][-window:]
        chosen_cum = past_cum.loc[(past_cum.index.month == cpi_cum_r.index[i].month)][-window:].dropna()
        if len(chosen) > 0:
            avg = chosen.mean()
            supseason_ind[i] = data[i] - avg
        if len(chosen_cum) > 0:
            avg_cum = chosen_cum.mean()
            supseason_cum_ind[i] = cpi_cum_r[i] - avg_cum
    return supseason_ind, supseason_cum_ind


def mean_Jan_Feb(ts):
    """Jan&Feb data Adjustment
    将一个序列的1,2月份的数据进行平均处理
    1,2月的数据都取1,2月数据的均值
    """
    # 输入的时间序列中1月和2月数据必须成对出现！
    ts_ave12 = ts.copy()
    Jan = pd.date_range(ts.index[0], ts.index[-1], freq='A-JAN')
    Feb = pd.date_range(ts.index[0], ts.index[-1], freq='A-FEB')
    ave12 = (ts.loc[Jan].values + ts.loc[Feb].values) / 2  # 将1,2月的数据进行修正
    ts_ave12.loc[Jan] = ave12
    ts_ave12.loc[Feb] = ave12
    ts_ave12.name = 'mean_Jan_Feb'

    return ts_ave12


def get_idx(yoyr, yoycr, base, base_year=2011):
    """
    输入同比和累计同比数据，并制定记住年，获取定基指数
    输入为当月同比，累计同比，基值，基年（默认2015年）,因为CPI和IP的wind提供的定基指数的时长都不长，
    因此综合考虑，我们觉得放在2015年是比较合适的。
    :param yoyr: 当月同比数据
    :param yoycr: 累计同比数据
    :param base: 原始定基指数
    :param base_year: 定基年，默认2015
    :return:
    """
    # 初始化
    date_index = pd.date_range(pd.to_datetime('1990-01-31'), yoycr.index[-1], freq='M')
    idx = pd.DataFrame(index=date_index)
    Sec_Names = yoycr.columns

    for sec_name in Sec_Names:  # 将基年总工业增加值序列（12个月）设为各行业工业增加值的基准
        idx.loc[str(base_year), sec_name] = base[str(base_year)] / base[str(base_year)].iloc[0] * 100

    # 推算基年之后数据
    year = base_year + 1
    while pd.to_datetime(str(year) + '-01-31') <= yoycr.index[-1]:
        year_idx = yoycr[str(year)].index
        last_year_idx = year_idx - pd.tseries.offsets.MonthEnd() * 12
        idx.loc[year_idx] = idx.loc[last_year_idx].values * yoyr.loc[year_idx].values
        idx.loc[year_idx[1]] = idx.loc[last_year_idx[1]].values * yoycr.loc[year_idx[1]].values
        idx.loc[year_idx[0]] = idx.loc[year_idx[1]]  # 1月等于2月
        year = year + 1

    # 推算基年之前数据
    year = base_year
    while pd.to_datetime(str(year) + '-12-31') >= yoycr.index[0]:
        next_year_idx = yoycr[str(year)].index
        year_idx = next_year_idx - pd.tseries.offsets.MonthEnd() * 12
        idx.loc[year_idx] = idx.loc[next_year_idx].values / yoyr.loc[next_year_idx].values
        if pd.to_datetime(str(year - 1) + '-01-31') == year_idx[0]:  # 去年的一月份
            # 2月份的数据用累计同比倒推，并使得1月等于2月
            idx.loc[year_idx[1]] = idx.loc[next_year_idx[1]].values / yoycr.loc[next_year_idx[1]].values
            idx.loc[year_idx[0]] = idx.loc[year_idx[1]]  # 1月等于2月
        year = year - 1

    return idx


def cum2single(series, missingJan=False, pctChange=False):
    """
    将累计值转为单月值 input: pd.series/dataframe
    # 若1月数据缺失 missingJan = True
    # 函数默认返回当月值 要返回环比 pctChange = True
    :param series:
    :param missingJan:
    :param pctChange:
    :return:
    """
    monthindex = pd.date_range(start=series.index[0], end=series.index[-1], freq='A-JAN')
    if missingJan:
        monthindex = pd.date_range(start=series.index[0], end=series.index[-1], freq='A-FEB')
    new_series = series.diff()
    new_series.loc[monthindex] = series.loc[monthindex]
    if pctChange:
        lenzero = len(new_series.loc[new_series.iloc[:, 0] == 0])
        print('caution: number of zeros in series: {0!s}, which may lead NaN or Inf'.format(lenzero))
        new_series = new_series.pct_change()
    return new_series


"""
#-#-#-#-#-#-#-#-#-#-#-#-#-#-季节性调整-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-

所有输入应当是月频时间序列，即以月频datetime为index的pandas.Series
需要下载的模组：numpy，pandas，scipy
关于季节性调整的资料参考以下网址：
http://cn.mathworks.com/help/econ/seasonal-adjustment-using-snxd7m-seasonal-filters.html

#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#--#-#-#-#-#-
"""

import numpy as np
import pandas as pd
import scipy.signal as sg


# -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-滤波函数#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-

def mean_Jan_Feb(ts):
    """
    平均1、2月数据
    输入的时间序列中1月和2月数据必须成对出现！
    :param ts:
    :return:
    """
    ts_ave12 = ts.copy()
    Jan = pd.date_range(ts.index[0], ts.index[-1], freq='A-JAN')
    Feb = pd.date_range(ts.index[0], ts.index[-1], freq='A-FEB')
    ave12 = (ts.loc[Jan].values + ts.loc[Feb].values) / 2
    ts_ave12.loc[Jan] = ave12
    ts_ave12.loc[Feb] = ave12
    ts_ave12.name = 'mean_Jan_Feb'

    return ts_ave12


def MA13(ts):  # 13-Term Moving Average

    # Symmetric weights for 13-Term MA
    sw13 = np.zeros(13) + 1 / 12
    sw13[0] = 1 / 24
    sw13[-1] = 1 / 24

    arr_ts_MA13 = sg.convolve(ts, sw13, 'same')  # 卷积
    # the following two lines deals with data at end of Series
    arr_ts_MA13[0:6] = arr_ts_MA13[6]
    arr_ts_MA13[-6:] = arr_ts_MA13[-7]
    ts_MA13 = pd.Series(arr_ts_MA13, index=ts.index)
    ts_MA13.name = '13-Term Moving Average'

    return ts_MA13


def MA13C(ts):
    """
    13-Term Moving Average for centering data
    :param ts:
    :return:
    """
    # Symmetric weights for 13-Term MA
    sw13 = np.zeros(13) + 1 / 12
    sw13[0] = 1 / 24
    sw13[-1] = 1 / 24

    arr_ts_MA13C = sg.convolve(ts, sw13, 'same')
    # the following two lines deals with data at end of Series
    # different from original MA13
    arr_ts_MA13C[0:6] = arr_ts_MA13C[12:18]
    arr_ts_MA13C[-6:] = arr_ts_MA13C[-18:-12]
    ts_MA13C = pd.Series(arr_ts_MA13C, index=ts.index)
    ts_MA13C.name = '13-Term Moving Average C'

    return ts_MA13C


def H13(ts):  # 13-Term Henderson filter

    # Henderson filter weights
    swh = np.array([-0.019, -0.028, 0, .066, .147, .214,
                    .24, .214, .147, .066, 0, -0.028, -0.019])
    # Asymmetric weights for end of series
    awh = np.array([[-.034, -.017, .045, .148, .279, .421],
                    [-.005, .051, .130, .215, .292, .353],
                    [.061, .135, .201, .241, .254, .244],
                    [.144, .205, .230, .216, .174, .120],
                    [.211, .233, .208, .149, .080, .012],
                    [.238, .210, .144, .068, .002, -.058],
                    [.213, .146, .066, .003, -.039, -.092],
                    [.147, .066, .004, -.025, -.042, 0],
                    [.066, .003, -.020, -.016, 0, 0],
                    [.001, -.022, -.008, 0, 0, 0],
                    [-.026, -.011, 0, 0, 0, 0],
                    [-.016, 0, 0, 0, 0, 0]])

    # Apply 13-term Henderson filter
    arr_h13 = sg.convolve(ts, swh, 'same')
    arr_h13[-6:] = sg.convolve(ts[-12:].values[:, None], awh, 'valid')
    arr_h13[0:6] = sg.convolve(ts[0:12].values[:, None], np.rot90(awh, 2), 'valid')
    h13 = pd.Series(arr_h13, index=ts.index)
    h13.name = '13-Term Henderson Filter'

    return h13


def S33(ts):  # S3x3 seasonal components

    T = len(ts)
    sidx = []
    for i in range(12):
        sidx.append(np.arange(i, T, 12))

    # Symmetric weights
    sw3 = np.array([1 / 9, 2 / 9, 1 / 3, 2 / 9, 1 / 9])
    # Asymmetric weights for end of series
    aw3 = np.array([[.259, .407], [.37, .407], [.259, .185], [.111, 0]])
    aw3R = [[0, .111], [.185, .259], [.407, .37], [.407, .259]]
    # Apply filter to each month
    shat3 = ts.copy()
    for i in range(12):
        ns = len(sidx[i])
        dat = ts[sidx[i]].copy()
        sd = sg.convolve(dat.values, sw3, 'same')
        sd[0:2] = sg.convolve(dat[0:4].values[:, None], np.rot90(aw3, 2), 'valid')
        sd[-2:] = sg.convolve(dat[-4:].values[:, None], aw3, 'valid')
        shat3.iloc[sidx[i]] = sd

    # 13-term moving average of filtered series
    sb = MA13C(shat3)
    # Center to get final estimate
    s33 = shat3 / sb
    s33.name = 'Seasonal component estimates'

    return s33


def S35(ts):  # S3x5 seasonal Components

    T = len(ts)
    sidx = []
    for i in range(12):
        sidx.append(np.arange(i, T, 12))

    # Symmetric weights
    sw5 = np.zeros(7) + 1 / 5
    sw5[0:2] = [1 / 15, 2 / 15]
    sw5[-2:] = [2 / 15, 1 / 15]
    # Asymmetric weights for end of series
    aw5 = np.array([[.150, .250, .293],
                    [.217, .250, .283],
                    [.217, .250, .283],
                    [.217, .183, .150],
                    [.133, .067, 0],
                    [.067, 0, 0]])
    # Apply filter to each month
    shat5 = ts.copy()
    for i in range(12):
        ns = len(sidx[i])
        dat = ts[sidx[i]].copy()
        sd = sg.convolve(dat.values, sw5, 'same')
        sd[0:3] = sg.convolve(dat[0:6].values[:, None], np.rot90(aw5, 2), 'valid')
        sd[-3:] = sg.convolve(dat[-6:].values[:, None], aw5, 'valid')
        shat5.iloc[sidx[i]] = sd

    # 13-term moving average of filtered series
    sb = MA13C(shat5)
    # Center to get final estimate
    s35 = shat5 / sb
    s35.name = 'Seasonal component estimates'

    return s35


# -#-#-#-#-#-#-#-derive deseasonalized series-#-#-#-#-#-#-#-

def Deseason(ts, Jan_Feb=False):
    """
    # -----是否平均1、2月数据-----
    :param ts:
    :param Jan_Feb:
    :return:
    """
    if Jan_Feb:  # 如果选择平均，则对1,2月数据进行取平均，否则就不做处理
        IVA_av12 = mean_Jan_Feb(ts)
    else:
        IVA_av12 = ts

    # -----Detrend the data using a 13-term moving average-----
    IVA_MA13 = MA13(IVA_av12)
    IVA_xt = IVA_av12 / IVA_MA13

    # -----Apply an S(3,3) seasonal filter to deseasonalize-----
    IVA_S33 = S33(IVA_xt)
    IVA_DeSeason = IVA_av12 / IVA_S33  # Deseasonalize series

    # -----Apply a 13-term Henderson filter to improve detrending-----
    IVA_H13 = H13(IVA_DeSeason)
    IVA_xt2 = IVA_av12 / IVA_H13  # New detrended series

    # -----Apply an S(3,5) seasonal filter to improve deseasonalizing-----
    IVA_S35 = S35(IVA_xt2)
    IVA_DeSeason2 = IVA_av12 / IVA_S35
    IVA_DeSeason2.name = 'Deseasonalized IVA'

    return IVA_DeSeason2

################### X13 ########################



from pyfi import WindHelper as w, line_graph, double_lines
from pyfi.common import Map
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

"""
performace
indicator
"""
maxdd = lambda eqd: (eqd.cumsum().expanding().max() - eqd.cumsum()).max()


def m_sharpe(eqd, bk=0.0):
    """monthly sharpe ratio"""
    try:
        if eqd.sum() == 0:
            return -100
        return ((eqd.mean() - bk / 12) / eqd.std()) * (12 ** 0.5)
    except BaseException as e:
        print(e)


def excess_msp(eqd, base_eqd, bk=0.0):
    return m_sharpe(eqd, bk) - m_sharpe(base_eqd, bk)


EQUITY = "000300.SH"  # 沪深300
FIXEDINCOME = "CBA01552.CS"  # 净价指数


def delta_money_ind(score):
    begin_date = score.index[0]
    end_date = score.index[-1]
    r007 = w.edb(codes=["M0041653"], begin_date=begin_date, end_date=end_date).iloc[:, 0].resample("M").last()
    d_money_ind = (r007 / 100 / 12).shift(1).dropna()
    return d_money_ind


def excess_msp(eqd, base_eqd, bk=0.0):
    return m_sharpe(eqd, bk) - m_sharpe(base_eqd, bk)


def monthly_backtest(score, pattern=1, title="Monthly backtest with net value index", weight_bounds=[-1, 1],
                     bk_code=FIXEDINCOME):
    """
    分值为发布时间点,也就是建仓点,基于净价指数进行择时
    :param title: 回测模板的名称
    :param score:
    :param pattern: 1: [0,1,2]; 5: [0,0.2,0.4,0.6,0.8,1,1.2,1.4,1.6,1.8,2.0]
    :return:
    """
    init_weight = (weight_bounds[0] + weight_bounds[1]) / 2.0
    low_weight = weight_bounds[0]
    up_weight = weight_bounds[1]
    begin_date = score.index[0]
    end_date = score.index[-1]
    ytm = w.edb(codes=["gz10y"], begin_date=begin_date, end_date=end_date).iloc[:, 0]
    bond_ind = w.wsd(code=bk_code, begin_date=begin_date, end_date=end_date, paras=["close"]).iloc[:, 0].resample(
        "M").last()  # 1结尾为财富值， 2结尾为净价
    #    bond_ind = W.edb(["M0265833"], begin_date, end_date).iloc[:,0].resample("M").last()
    # 仓位设计部分开始
    if pattern > 1:
        pos = score.apply(lambda x: init_weight + (up_weight - low_weight) / 2.0 / pattern * x).shift(1)
    elif pattern == 1:
        pos = score.apply(lambda x: low_weight if x < 0 else (init_weight if x == 0 else up_weight)).shift(1)
    else:
        raise Exception("The pattern is invalid, man ~~ please reset it, OK??? Only 1 and 5 is permitted")
    # 仓位设计部分结束
    d_bond_ind = bond_ind.pct_change().dropna()
    benchmark = (1 + init_weight * d_bond_ind).cumprod()
    # pos_ = WindHelper.monthly_data_with_td(pos)
    port = (pos * d_bond_ind + 1).cumprod().dropna()
    alpha = port - benchmark
    # 看多时候的alpha & 看空时候的alpha
    d_alpha = alpha.diff()
    d_alpha[0] = alpha[0]
    alpha_long = d_alpha[pos > init_weight]  # pos不存在计算误差
    alpha_short = d_alpha[pos < init_weight]
    length_alpha_long = len(alpha_long) * 1.0  # 所有做多的时点数
    length_alpha_short = len(alpha_short) * 1.0  # 所有做空的时点数
    length_long_precise = len(alpha_long[alpha_long > 0]) * 1.0  # 持仓不同，一般不会存在误差，除非资产价格不动。
    length_short_precise = len(alpha_short[alpha_short > 0]) * 1.0
    # draw the graph
    fig = plt.figure(figsize=(10, 15))
    ax1 = plt.subplot(311)
    double_lines(series1=score, series2=ytm, title="ytm V.S. score",
                 lgd1="score", lgd2="10y ytm", ax=ax1, fig=fig)
    ax2 = plt.subplot(312)
    ax3 = ax2.twinx()
    line_graph([port, benchmark], legend_list=["port", "benchmark"],
               title="curve battle", ax=ax2, fig=fig)
    ax3.fill_between(alpha.index, 0, alpha.values, color="grey", alpha=0.3)
    ax4 = plt.subplot(313)
    line_graph([pos], legend_list=["position"],
               title="Check the position change frequencey", ax=ax4, fig=fig)
    plt.suptitle(title, fontsize=36)
    plt.show()
    d_port = port.pct_change().dropna()
    d_bench = benchmark.pct_change().dropna()
    port_yield = (port[-1] ** (1 / ((len(port)) / 12))) - 1
    benchmark_yield = (benchmark[-1] ** (1 / ((len(benchmark)) / 12))) - 1
    return Map({"score": score,
                "port": port,
                "pos": pos,
                "benchmark": benchmark,
                "alpha": alpha,
                "d_alpha": d_alpha,
                "ytm": ytm.resample("M").last(),
                "report": Map({
                    "最大回撤": maxdd(d_port),
                    "夏普比率": m_sharpe(d_port),
                    "超额夏普比率": excess_msp(d_port, d_bench),
                    "年化收益率": (port[-1] - 1) / ((len(port)) / 12),
                    "基准年化收益率": benchmark_yield,
                    "年均alpha": port_yield - benchmark_yield,
                    "alpha最大回撤": maxdd(d_alpha),
                    "信息比率": (port_yield - benchmark_yield) / (d_port.sub(d_bench).std() * np.sqrt(12)),
                    "calmar指标": (port[-1] ** (1 / (len(port) - 1) * 12) - 1) / maxdd(d_port),
                    "看多次数": length_alpha_long,
                    "看多准确率": np.nan if length_alpha_long == 0 else length_long_precise / length_alpha_long,
                    "看空次数": length_alpha_short,
                    "看空准确率": np.nan if length_alpha_short == 0 else length_short_precise / length_alpha_short,
                    "综合准确率": np.nan if (length_alpha_long + length_alpha_short) == 0 else (
                                                                                                  length_long_precise + length_short_precise) / (
                                                                                                  length_alpha_long + length_alpha_short),
                })
                })


def monthly_backtest_with_repo(score, pattern=1, init_weight=1):
    """
    分值为发布时间点,也就是建仓点
    """
    b_weight = init_weight
    import matplotlib.pyplot as plt
    begin_date = score.index[0]
    end_date = score.index[-1]
    if pattern > 1:
        pos = score.apply(lambda x: b_weight + (b_weight / pattern) * x).shift(1)
    elif pattern == 1:
        pos = score.apply(lambda x: 0 if x < 0 else (b_weight if x == 0 else 2 * b_weight)).shift(1)
    else:
        raise Exception("The pattern is invalid, man ~~ please reset it, OK??? Only 1 and 5 is permitted")
    ytm = w.edb(codes=["gz10y"], begin_date=begin_date, end_date=end_date).iloc[:, 0]
    # r007 = w.edb(codes=["M0041653"], begin_date=begin_date, end_date=end_date).iloc[:, 0].resample("M").last()
    bond_ind = w.wsd("CBA01551.CS", ["close"], begin_date, end_date).iloc[:, 0] \
        .resample("M").last()  # 1结尾为财富值， 2结尾为净价
    d_bond_ind = bond_ind.pct_change().dropna()
    r007 = w.edb(codes=["M0041653"], begin_date=begin_date, end_date=end_date).iloc[:, 0].resample("M").last()
    # money_ind = (r007 / 100 / 12 + 1).cumprod().shift(1).dropna()
    d_money_ind = (r007 / 100 / 12).shift(1).dropna()
    # establish the benchmark
    benchmark = (d_bond_ind * b_weight + (d_money_ind * (1 - b_weight)) + 1).cumprod()
    # pos_ = WindHelper.monthly_data_with_td(pos)
    port = (pos * d_bond_ind + (1 - pos) * d_money_ind + 1).cumprod().dropna()
    alpha = port - benchmark
    # 看多时候的alpha & 看空时候的alpha
    d_alpha = alpha.diff()
    d_alpha[0] = alpha[0]
    alpha_long = d_alpha[pos > init_weight]  # pos不存在计算误差
    alpha_short = d_alpha[pos < init_weight]
    length_alpha_long = len(alpha_long) * 1.0  # 所有做多的时点数
    length_alpha_short = len(alpha_short) * 1.0  # 所有做空的时点数
    length_long_precise = len(alpha_long[alpha_long > 0]) * 1.0  # 持仓不同，一般不会存在误差，除非资产价格不动。
    length_short_precise = len(alpha_short[alpha_short > 0]) * 1.0
    # draw the picture
    fig = plt.figure(figsize=(10, 15))
    ax1 = plt.subplot(311)
    double_lines(series1=score, series2=ytm, title="ytm V.S. score",
                 lgd1="score", lgd2="10y ytm", ax=ax1, fig=fig)
    ax2 = plt.subplot(312)
    ax3 = ax2.twinx()
    line_graph([port, benchmark], legend_list=["port", "benchmark"],
               title="curve battle", ax=ax2, fig=fig)
    ax3.fill_between(alpha.index, 0, alpha.values, color="grey", alpha=0.3)
    ax4 = plt.subplot(313)
    line_graph([pos], legend_list=["position"],
               title="Check the position change frequencey", ax=ax4, fig=fig)
    plt.suptitle("Monthly Backtest with wealth index", fontsize=36)
    plt.show()
    d_port = port.pct_change().dropna()
    d_bench = benchmark.pct_change().dropna()
    port_yield = (port[-1] ** (1 / ((len(port)) / 12))) - 1
    benchmark_yield = (benchmark[-1] ** (1 / ((len(benchmark)) / 12))) - 1
    return Map({"score": score,
                "port": port,
                "pos": pos,
                "benchmark": benchmark,
                "alpha": alpha,
                "ytm": ytm.resample("M").last(),
                "report": Map({
                    "最大回撤": maxdd(d_port),
                    "夏普比率": m_sharpe(d_port),
                    "超额夏普比率": excess_msp(d_port, d_bench),
                    "年化收益率": port_yield,
                    "基准年化收益率": benchmark_yield,
                    "年均alpha": port_yield - benchmark_yield,
                    "alpha最大回撤": maxdd(d_alpha),
                    "信息比率": (d_port.sub(d_bench)).mean() / (d_port.sub(d_bench).std()),
                    "calmar指标": (port[-1] ** (1 / (len(port) - 1) * 12) - 1) / maxdd(d_port),
                    "看多次数": length_alpha_long,
                    "看多准确率": np.nan if length_alpha_long == 0 else length_long_precise / length_alpha_long,
                    "看空次数": length_alpha_short,
                    "看空准确率": np.nan if length_alpha_short == 0 else length_short_precise / length_alpha_short,
                    "综合准确率": np.nan if (length_alpha_long + length_alpha_short) == 0 else (
                                                                                                  length_long_precise + length_short_precise) / (
                                                                                                  length_alpha_long + length_alpha_short),

                })
                })


def monthly_backtest_with_repo2(score, pattern=1):
    """
    保持仓位始终不变，防止前期的alpha对后期的alpha带来增益的影响
    分值为发布时间点,也就是建仓点

    """
    b_weight = 1
    import matplotlib.pyplot as plt
    begin_date = score.index[0]
    end_date = score.index[-1]
    if pattern == 5:
        pos = score.apply(lambda x: b_weight + (b_weight / pattern) * x).shift(1)
    elif pattern == 1:
        pos = score.apply(lambda x: 0 if x < 0 else (1 if x == 0 else 2)).shift(1)
    else:
        raise Exception("The pattern is invalid, man ~~ please reset it, OK??? Only 1 and 5 is permitted")
    ytm = w.edb(codes=["gz10y"], begin_date=begin_date, end_date=end_date).iloc[:, 0]
    r007 = w.edb(codes=["M0041653"], begin_date=begin_date, end_date=end_date).iloc[:, 0].resample("M").last()
    bond_ind = w.wsd("CBA01551.CS", ["close"], begin_date, end_date).iloc[:, 0] \
        .resample("M").last()  # 1结尾为财富值， 2结尾为净价
    d_bond_ind = bond_ind.pct_change().dropna()
    r007 = w.edb(codes=["M0041653"], begin_date=begin_date, end_date=end_date).iloc[:, 0].resample("M").last()
    # money_ind = (r007 / 100 / 12 + 1).cumprod().shift(1).dropna()
    d_money_ind = (r007 / 100 / 12).shift(1).dropna()
    # establish the benchmark
    benchmark = 1 + (d_bond_ind * b_weight + (d_money_ind * (1 - b_weight))).cumsum()
    # pos_ = WindHelper.monthly_data_with_td(pos)
    port = 1 + (pos * d_bond_ind + (1 - pos) * d_money_ind).cumsum().dropna()
    alpha = port - benchmark
    # draw the picture
    fig = plt.figure(figsize=(10, 15))
    ax1 = plt.subplot(311)
    double_lines(series1=score, series2=ytm, title="ytm V.S. score",
                 lgd1="score", lgd2="10y ytm", ax=ax1, fig=fig)
    ax2 = plt.subplot(312)
    ax3 = ax2.twinx()
    line_graph([port, benchmark], legend_list=["port", "benchmark"],
               title="curve battle", ax=ax2, fig=fig)
    ax3.fill_between(alpha.index, 0, alpha.values, color="grey", alpha=0.3)
    ax4 = plt.subplot(313)
    line_graph([pos], legend_list=["position"],
               title="Check the position change frequencey", ax=ax4, fig=fig)
    plt.suptitle("Monthly Backtest with wealth index(No cum)", fontsize=36)
    plt.show()
    return {"score": score,
            "port": port,
            "pos": pos,
            "benchmark": benchmark,
            "alpha": alpha,
            "ytm": ytm.resample("M").last()
            }


def monthly_backtest_with_structure(score,
                                    struct_scores=None,
                                    pattern=1, init_weight=0.5,
                                    init_duration_allocation=[0.25, 0.25, 0.25, 0.25]):
    """
    输入久期打分，输入品种打分，输入模式，输入初始权重,基准权重
    注意：score的最后一个分值，是不会考虑到回测中的
    分值越高，权重越大
    :param score: 久期择时打分
    :param struct_scores: 期限结构打分，DataFrame 格式, 0-5分
    :param pattern: 调仓模式
    :param init_weight: 中性仓位
    :param init_duration_allocation：基准的久期的权重分配
    :return:
    """
    # 输入参数校验
    # 确定基准久期结构
    benchmark_struct = pd.DataFrame([init_duration_allocation] * len(score), index=score.index,
                                    columns=["3y", "5y", "7y", "10y"])
    if struct_scores is not None:
        if len(score) > len(struct_scores):
            struct_scores = pd.DataFrame(struct_scores, index=score.index)
            for i in range(len(score)):
                if struct_scores.iloc[i, :].isnull().any():
                    struct_scores.iloc[i, :] = np.array(init_duration_allocation)

    else:
        struct_scores = benchmark_struct
    # 确定久期
    begin_date = score.index[0]
    end_date = score.index[-1]

    one_to_three = w.wsd("CBA01521.CS", ["close", "duration"], begin_date=begin_date, end_date=end_date) \
        .resample("M").last()  # 1结尾为财富值， 2结尾为净价
    three_to_five = w.wsd("CBA01531.CS", ["close", "duration"], begin_date=begin_date, end_date=end_date) \
        .resample("M").last()  # 1结尾为财富值， 2结尾为净价
    five_to_seven = w.wsd("CBA01541.CS", ["close", "duration"], begin_date=begin_date, end_date=end_date) \
        .resample("M").last()  # 1结尾为财富值， 2结尾为净价
    seven_to_ten = w.wsd("CBA01551.CS", ["close", "duration"], begin_date=begin_date, end_date=end_date) \
        .resample("M").last()  # 1结尾为财富值， 2结尾为净价
    # 资金净值曲线
    r007 = w.edb(codes=["M0041653"], begin_date=begin_date, end_date=end_date).iloc[:, 0].resample("M").last()
    # money_ind = (r007 / 100 / 12 + 1).cumprod().shift(1).dropna()
    d_money_ind = (r007 / 100 / 12).shift(1).dropna()
    # 十年国债收益率
    ytm = w.edb(codes=["gz10y"], begin_date=begin_date, end_date=end_date).iloc[:, 0]
    # 子指数的每个月的收益率
    d_one_to_three = one_to_three.close.pct_change().dropna()  # shift1
    d_three_to_five = three_to_five.close.pct_change().dropna()  # shift1
    d_five_to_seven = five_to_seven.close.pct_change().dropna()  # shift1
    d_seven_to_ten = seven_to_ten.close.pct_change().dropna()  # shift1
    # 生成基准中的债券组合部分
    bond_ind = pd.Series(index=score.index)  # shift0
    bond_duration = pd.Series(index=score.index)  # shift0
    for i in range(len(bond_ind)):
        if i == 0:
            bond_ind[0] = 1.0
            continue
        tmp_struct_durations = np.array([one_to_three.duration[i - 1],  # 最后一个久期数据无效
                                         three_to_five.duration[i - 1],
                                         five_to_seven.duration[i - 1],
                                         seven_to_ten.duration[i - 1]])
        d_tmp_struct = np.array([d_one_to_three[i - 1],
                                 d_three_to_five[i - 1],
                                 d_five_to_seven[i - 1],
                                 d_seven_to_ten[i - 1]])
        lmda = 1.0 / ((np.array(init_duration_allocation) / tmp_struct_durations).sum())
        tmp_struct_weights = np.array(init_duration_allocation) / tmp_struct_durations * lmda
        bond_ind[i] = (1 + (tmp_struct_weights * d_tmp_struct).sum()) * bond_ind[i - 1]
        bond_duration[i - 1] = (tmp_struct_weights * tmp_struct_durations).sum()  # shfit0
    d_bond_ind = bond_ind.pct_change().dropna()  # shift1
    # 构建加入资金权重的基准
    benchmark = (d_bond_ind * init_weight + (d_money_ind * (1 - init_weight)) + 1).cumprod()
    if pattern == 5:
        pos = score.apply(lambda x: init_weight + (init_weight / pattern) * x).shift(0)  # shift0坐标
    elif pattern == 1:
        pos = (score.apply(lambda x: 0 if x < 0 else (1 if x == 0 else 2)) * init_weight).shift(0)  # shift0坐标
    else:
        raise Exception("The pattern is invalid, man ~~ please reset it, OK??? Only 1 and 5 is permitted")
    port_duration = bond_duration * pos  # shift0
    pos = pos.shift(1).dropna()
    # port1 无结构优化
    port1 = (pos * d_bond_ind + (1 - pos) * d_money_ind + 1).cumprod().dropna()  # 都会shift1
    alpha1 = port1 - benchmark

    if struct_scores is not None:
        struct_scores = struct_scores.loc[(struct_scores.index >= begin_date) & (struct_scores.index <= end_date),
                        :].shift(1).dropna()
        struct_pos = pd.DataFrame(columns=struct_scores.columns, index=pos.index)
        struct_durations = pd.DataFrame(columns=struct_scores.columns, index=port_duration.index)  # shift0
        # port2 结构优化
        port2 = pd.Series(index=port1.index)  # shift1
        d_port2 = pd.Series(index=port1.index)  # shift1
        try:
            for i in range(len(port2)):
                # 确定当天的无结构优化组合的久期
                tmp_duration = port_duration[i]  # 标量, shift0
                # 确定给整体比率
                lmda = tmp_duration / (struct_scores.iloc[i, :].sum())  # 标量， struct_scores已经shift, 注意这里是shift1
                tmp_struct_durations = np.array([one_to_three.duration[i],
                                                 three_to_five.duration[i],
                                                 five_to_seven.duration[i],
                                                 seven_to_ten.duration[i]])  # shift0
                struct_durations.iloc[i, :] = tmp_struct_durations  # N*4 shift0
                tmp_struct_weights = (struct_scores.iloc[i, :] * lmda) / tmp_struct_durations
                tmp_pos = tmp_struct_weights.sum()  # 标量
                # print(tmp_pos)
                struct_pos.iloc[i, :] = tmp_struct_weights  # shift1
                bond_return = (tmp_struct_weights * np.array([d_one_to_three[i],
                                                              d_three_to_five[i],
                                                              d_five_to_seven[i],
                                                              d_seven_to_ten[i]])).sum()  # 实际计算的是shift1的第一个时点
                port_return = bond_return + (1 - tmp_pos) * d_money_ind[i]
                d_port2[i] = port_return
                if i == 0:
                    port2[i] = 1 * (port_return + 1)
                else:
                    port2[i] = port2[i - 1] * (port_return + 1)
        except IndexError as e:
            print(pos.index[i])
            raise e
        alpha2 = port2 - benchmark
        port = port2
    else:
        alpha2 = alpha1
        port = port1
        port2 = port1
    alpha = port - benchmark
    # 开始画图
    fig = plt.figure(figsize=(10, 15))
    ax1 = plt.subplot(311)
    double_lines(series1=score, series2=ytm, title="ytm V.S. score",
                 lgd1="score", lgd2="10y ytm", ax=ax1, fig=fig)
    ax2 = plt.subplot(312)
    ax3 = ax2.twinx()
    line_graph([port1, port2, benchmark], legend_list=["duration", "structural duration", "benchmark"],
               title="curve battle", ax=ax2, fig=fig)
    ax3.fill_between(alpha.index, 0, alpha.values, color="grey", alpha=0.3)
    ax4 = plt.subplot(313)
    line_graph([pos], legend_list=["position"],
               title="Check the position change frequencey", ax=ax4, fig=fig)
    plt.suptitle("Monthly Backtest with wealth index", fontsize=36)
    plt.show()
    d_port = port.pct_change().dropna()
    d_bench = benchmark.pct_change().dropna()
    return Map({"score": score,
                "port1": port1,
                "port2": port2,
                "pos": pos,
                "struct_pos": struct_pos,
                "struct_durations": struct_durations,
                "port_duration": port_duration,
                "benchmark": benchmark,
                "alpha": alpha,
                "ytm": ytm.resample("M").last(),
                "report": Map({
                    "最大回撤": maxdd(d_port),
                    "夏普比率": m_sharpe(d_port),
                    "最终alpha1": alpha1[-1],
                    "最终alpha2": alpha2[-1],
                    "久期结构alpha": alpha2[-1] - alpha1[-1],
                    "超额夏普比率": excess_msp(d_port, d_bench),
                    "年均alpha": alpha[-1] / ((len(port) - 1) / 12),
                    "calmar指标": (port[-1] ** (1 / (len(port) - 1) * 12) - 1) / maxdd(d_port),
                })
                })



import math
import numpy as np
import pandas as pd


class BootstrapYieldCurve(object):
    def __init__(self, ytms=[]):
        self.terms = list(range(1, len(ytms) + 1))
        self.ytms = pd.Series(ytms, index=self.terms)
        self.spot_rates = pd.Series(index=self.terms)
        self.forward_rates = pd.Series(index=self.terms)

    def get_ytms(self):
        """
        获取到期收益率曲线
        :return:
        """
        return self.ytms

    def get_spot_rates(self):
        """
        获取即期收益率曲线
        :return:
        """
        for T in self.terms:
            ytm = self.ytms.loc[T]
            self.spot_rates.loc[T] = self.__calculate_spot_rate__(T, ytm)
        return self.spot_rates

    def get_forward_rates(self, key_terms=[]):
        """
        获取远期收益率曲线
        :return:
        """
        if len(key_terms) == 0:
            # 无期限筛选
            key_terms = self.terms
        self.spot_rates = self.get_spot_rates()
        self.forward_rates = pd.Series(index=key_terms)
        self.forward_rates.iloc[0] = self.spot_rates.iloc[0]
        for i in range(1, len(key_terms)):
            t = key_terms[i]
            t_ahead = key_terms[i - 1]
            self.forward_rates.iloc[i] = 100 * ((((self.spot_rates.loc[t] / 100 + 1) ** t) / (
                        (self.spot_rates.loc[t_ahead] / 100 + 1) ** t_ahead))**(1/(t-t_ahead))-1)
        return self.forward_rates

    def __calculate_spot_rate__(self, T, ytm):
        value = 100
        for i in range(T - 1):
            t = i + 1
            spot_rate = self.spot_rates.iloc[i]
            discounted_coupon = ytm / ((1 + spot_rate / 100) ** t)
            value -= discounted_coupon
        spot_rate = (((ytm + 100) / value) ** (1 / T) - 1) * 100
        return spot_rate


if __name__ == "__main__":
    yield_curve = BootstrapYieldCurve(ytms=[3, 3.3, 3.4, 3.5, 3.6, 3.7])
    y = yield_curve.get_forward_rates()

    import matplotlib.pyplot as plt

    plt.plot(y.index, y.values)
    plt.title("Spot Rate Curve")
    plt.ylabel("Spot Rate (%)")
    plt.xlabel("Spot rates in Years")
    plt.show()



import pandas as pd
"""
TODO: 本部分代码为临时模块，未来会逐步归并到processing模块，
当前的主要功能是：
1. 获取定基指数
2. 1,2月数据进行处理
3. 将累计值转换为单月值
"""


def mean_Jan_Feb(ts):
    """Jan&Feb data Adjustment
    将一个序列的1,2月份的数据进行平均处理
    1,2月的数据都取1,2月数据的均值
    """
    # 输入的时间序列中1月和2月数据必须成对出现！
    ts_ave12 = ts.copy()
    Jan = pd.date_range(ts.index[0], ts.index[-1], freq='A-JAN')
    Feb = pd.date_range(ts.index[0], ts.index[-1], freq='A-FEB')
    ave12 = (ts.loc[Jan].values + ts.loc[Feb].values) / 2  # 将1,2月的数据进行修正
    ts_ave12.loc[Jan] = ave12
    ts_ave12.loc[Feb] = ave12
    ts_ave12.name = 'mean_Jan_Feb'

    return ts_ave12


def get_idx(yoyr, yoycr, base, base_year=2011):
    """
    输入同比和累计同比数据，并制定基准年，获取定基指数
    输入为当月同比，累计同比，基值，基年（默认2015年）,因为CPI和IP的wind提供的定基指数的时长都不长，
    因此综合考虑，我们觉得放在2015年是比较合适的。
    :param yoyr: 当月同比数据
    :param yoycr: 累计同比数据
    :param base: 原始定基指数
    :param base_year: 定基年，默认2015
    :return:
    """
    # 初始化
    date_index = pd.date_range(pd.to_datetime('1990-01-31'), yoycr.index[-1], freq='M')
    idx = pd.DataFrame(index=date_index)
    Sec_Names = yoycr.columns

    for sec_name in Sec_Names:  # 将基年总工业增加值序列（12个月）设为各行业工业增加值的基准
        idx.loc[str(base_year), sec_name] = base[str(base_year)] / base[str(base_year)].iloc[0] * 100

    # 推算基年之后数据
    year = base_year + 1
    while pd.to_datetime(str(year) + '-01-31') <= yoycr.index[-1]:
        year_idx = yoycr[str(year)].index
        last_year_idx = year_idx - pd.tseries.offsets.MonthEnd() * 12
        idx.loc[year_idx] = idx.loc[last_year_idx].values * yoyr.loc[year_idx].values
        idx.loc[year_idx[1]] = idx.loc[last_year_idx[1]].values * yoycr.loc[year_idx[1]].values
        idx.loc[year_idx[0]] = idx.loc[year_idx[1]]  # 1月等于2月
        year = year + 1

    # 推算基年之前数据
    year = base_year
    while pd.to_datetime(str(year) + '-12-31') >= yoycr.index[0]:
        next_year_idx = yoycr[str(year)].index
        year_idx = next_year_idx - pd.tseries.offsets.MonthEnd() * 12
        idx.loc[year_idx] = idx.loc[next_year_idx].values / yoyr.loc[next_year_idx].values
        if pd.to_datetime(str(year - 1) + '-01-31') == year_idx[0]:  # 去年的一月份
            # 2月份的数据用累计同比倒推，并使得1月等于2月
            idx.loc[year_idx[1]] = idx.loc[next_year_idx[1]].values / yoycr.loc[next_year_idx[1]].values
            idx.loc[year_idx[0]] = idx.loc[year_idx[1]]  # 1月等于2月
        year = year - 1

    return idx


def cum2single(series, missingJan=False, pctChange=False):
    """
    将累计值转为单月值 input: pd.series/dataframe
    # 若1月数据缺失 missingJan = True
    # 函数默认返回当月值 要返回环比 pctChange = True
    :param series:
    :param missingJan:
    :param pctChange:
    :return:
    """
    monthindex = pd.date_range(start=series.index[0], end=series.index[-1], freq='A-JAN')
    if missingJan:
        monthindex = pd.date_range(start=series.index[0], end=series.index[-1], freq='A-FEB')
    new_series = series.diff()
    new_series.loc[monthindex] = series.loc[monthindex]
    if pctChange:
        lenzero = len(new_series.loc[new_series.iloc[:, 0] == 0])
        print('caution: number of zeros in series: {0!s}, which may lead NaN or Inf'.format(lenzero))
        new_series = new_series.pct_change()
    return new_series



# encoding: utf-8
"""
提供WindPy接口的pandas封装
"""
from math import isnan
from datetime import timedelta
# app
from WindPy import w
import pandas as pd
from pyfi.mapper import label, da_mapper, mapper
from pyfi.data_agent import *
from pyfi.common import nearest


class WindHelper(object):
    """
    WindPy20170904更新
    在本次修改后, 下面函数结果中的时间Times将只包含日期date: 
    wsd, tdays, tdayscount, tdaysoffset, wset, weqs, wpd, htocode, edb
    常用字段：
    close
    settle
    volume
    """
    # 常见edb代码映射表
    """
    ip=工业增加值：当月同比
    cpi=CPI:当月同比(月）
    cpif=CPI食品：当月同比
    cpinf=CPI非食品：当月同比
    """
    mapper = mapper
    label = label
    da_mapper = da_mapper

    @classmethod
    def check_begin_and_end_date(cls):
        # TODO: 把时间校验模块化
        pass

    @classmethod
    def translate(cls, code):
        """
        将list中的code进行映射
        :param code:
        :return:
        """
        if type(code) == list:
            return [cls.mapper[x] if x in cls.mapper else x for x in code]
        else:
            if code in cls.mapper:
                return cls.mapper[code]
            else:
                return code

    @classmethod
    def wset(cls, *args, **kwargs):
        if not w.isconnected():
            w.start()
        return w.wset(*args, **kwargs)

    @classmethod
    def wsd(cls, code, paras, begin_date, end_date, options="credibility=1", adjust=False, shift=0, debug=True):
        """单代码多维日期序列
        或者多代码单维日期序列
        注意：wind并不支持多代码多维的数据提取
        :param adjust:
        :param shift:
        :return:
        :param code: (list) 一次只能一个品种
        :param paras: (list) fields
        :param begin_date: (datetime or str)如 '20150101'
        :param end_date: （datetime or str) 如 '20150101'
        :param options: （string） xx=xx;yy=yy 不支持大小写
        :return: pandas.DataFrame
        """
        if debug is True:
            print("wsd开始提取数据..." + str(code))
        if begin_date is None:
            raise Exception("codes:" + str(code) + "的begindate是空的！")
        if type(begin_date) is str:
            begin_date = datetime.strptime(begin_date, '%Y-%m-%d')
        if type(end_date) is str:
            end_date = datetime.strptime(end_date, '%Y-%m-%d')
        if begin_date > end_date:
            print("结束时间小于开始时间，你在逗我么，骚年！")
            return None
        # code = cls.mapper[code] if code in cls.mapper else code
        try:
            if not w.isconnected():
                w.start()
            # 校验，防止codes和paras同时为长度大于1的list
            if type(code) == list and type(paras) == list:
                if len(code) > 1 and len(paras) > 1:
                    raise Exception(u"wsd不能提取多代码多维度数据")
            if type(code) == list and len(code) == 1:
                code = code[0]
            code = cls.translate(code=code)
            options = options
            wind_data = w.wsd(code, paras, begin_date, end_date, options)
            if hasattr(wind_data, "ErrorCodes") and wind_data.ErrorCode != 0:
                print("wsd: wind 抛出异常！")
                raise (Exception(str(wind_data)))
            if wind_data is None:
                raise Exception(u"wsd调用wind服务端异常，数据为空")
            if len(wind_data.Data) == 0:
                raise Exception(u"wsd调用wind服务端异常，数据为空")
            if len(wind_data.Data[0]) == 0:
                raise Exception(u"wsd调用wind服务端异常，数据为空")
            if "CWSDService: invalid windcodes." in wind_data.Data[0]:
                raise Exception(wind_data.Data[0])
            if "CWSDService: No data." in wind_data.Data[0]:
                return None
            dataDict = {}
            if not hasattr(code, "__iter__") or type(code) is str:
                for i in range(len(wind_data.Data)):  # 单代码则按照因子索引
                    dataDict[wind_data.Fields[i].lower()] = wind_data.Data[i]
                    df = pd.DataFrame(dataDict, index=wind_data.Times)
            else:  # 多代码的情况下
                if begin_date == end_date:
                    df = pd.DataFrame(wind_data.Data, columns=code, index=wind_data.Times)  # 当只有一个日期时，返回的数据结构发生变化
                else:
                    for i in range(len(code)):
                        dataDict[code[i]] = wind_data.Data[i]
                        df = pd.DataFrame(dataDict, index=wind_data.Times)
                        df = df.reindex(columns=code)
            # 将date类型转换为datetime类型
            df.index = pd.to_datetime(df.index)
            df.index.name = "trade_date"

            # 二次调整
            if adjust:
                df = cls.monthly_data_with_td(df)
            if shift != 0:
                df = cls.monthly_data_with_signal(df, shift)
            return df
        except Exception as e:
            print(format(e))
            raise

    @classmethod
    def wss(cls, codes, paras):
        """多代码多维信息序列
        :param codes:(list) or (str)
        :param paras:(list) or (str)
        :return:(DataFrame);
        """

        try:
            if not w.isconnected():
                w.start()
            if type(codes) is not list:
                codes = codes.split(",")  # 全部转换为list格式
            codes = [cls.mapper[x] if x in cls.mapper else x for x in codes]
            if type(paras) is not list:
                paras = paras.split(",")
            paras.insert(0, "windcode")
            wind_data = w.wss(codes, paras, options=None)
            if hasattr(wind_data, "ErrorCodes") and wind_data.ErrorCode != 0:
                raise (Exception(str(wind_data)))
            if len(wind_data.Data) == 0:
                return None
            if len(wind_data.Data[0]) == 0:
                return None
            dataDict = {}
            for i in range(len(wind_data.Data)):
                dataDict[wind_data.Fields[i].lower()] = wind_data.Data[i]
            df = pd.DataFrame(dataDict)
            df = df[paras]
            df.rename(columns={"windcode": "code"}, inplace=True)

            return df
        except BaseException as e:
            print(format(e))
            raise

    @classmethod
    def edb(cls, codes, begin_date, end_date, options="fill=perious", adjust=False, shift=0, fill=False, debug=True):
        """多代码单维时间序列
        :param shift:
        :param fill:
        :param adjust: 是否对日期进行调整，td表示最近的交易日期代替月末数据
        :param codes:
        :param begin_date:
        :param end_date:
        :param options:
        :return:
        """
        if debug is True:
            print("edb开始提取数据..." + str(codes))
        if begin_date is None:
            raise Exception("codes:" + str(codes) + "的begindate是空的！")
        if type(begin_date) is str:
            begin_date = datetime.strptime(begin_date, '%Y-%m-%d')
        if type(end_date) is str:
            end_date = datetime.strptime(end_date, '%Y-%m-%d')
        if begin_date > end_date:
            print("结束时间小于开始时间，你在逗我么，骚年！")
            return None
        if not w.isconnected():
            w.start()
        if type(codes) is not list:
            codes = codes.split(",")  # 全部转换为list格式
        final_codes = []  # 因子名称列表
        final_data_mapper = {}  # 预处理函数字典
        for x in codes:
            if x in cls.da_mapper:
                final_data_mapper[x] = cls.da_mapper[x]
            else:
                final_codes.append(cls.mapper[x] if x in cls.mapper else x)
        try:
            df = pd.DataFrame()
            if len(final_codes) > 0:
                if not w.isconnected:
                    w.start()
                windData = w.edb(final_codes, begin_date, end_date, options)
                if hasattr(windData, "ErrorCodes") and windData.ErrorCode != 0:
                    raise (Exception(str(windData)))
                if len(windData.Data) == 0:
                    return None
                if len(windData.Data[0]) == 0:
                    return None
                if "Internet Timeout" in windData.Data[0]:
                    raise Exception("wind APi连接不上:" + windData.Data[0])
                dataDict = {}
                cols = []
                for i in range(len(windData.Data)):
                    col = windData.Codes[i]
                    if col in cls.mapper.values():
                        col = list(cls.mapper.keys())[list(cls.mapper.values()).index(col)]
                    cols.append(col)
                for i in range(len(windData.Data)):
                    # col = windData.Codes[i]
                    # if col in cls.mapper.values():
                    #     col = list(cls.mapper.keys())[list(cls.mapper.values()).index(col)]
                    dataDict[cols[i]] = windData.Data[i]
                df = df.join(pd.DataFrame(dataDict, index=windData.Times), how="outer")
                df.index = pd.to_datetime(df.index)
                df.index.name = "trade_date"
                df = df.loc[:, cols]  # 调整顺序
            # 自定义指标,直接通过其他封装函数获取
            if len(final_data_mapper) > 0:
                da_dfs = [(final_data_mapper[x])(begin_date, end_date) for x in final_data_mapper]
                da_df = da_dfs[0]
                for i in range(1, len(da_dfs)):
                    da_df = da_df.join(da_dfs[i], how="outer")
                df = df.join(da_df, how="outer")
            # 二次调整
            if fill:
                df = cls.fill(df, end_date)  # 对残缺的月份进行补全， 并用上一个数据进行补全
            if adjust:
                df = cls.monthly_data_with_td(df)
            if shift != 0:
                df = cls.monthly_data_with_signal(df, shift)
            return df
        except BaseException as e:
            print(format(e))
            raise

    @classmethod
    def wsi(cls, code, fields, trade_date, num_retries=2):
        """单代码多维"""
        try:
            if type(trade_date) in (datetime, pd._libs.tslib.Timestamp):
                trade_date = trade_date.strftime("%Y%m%d")
            if type(fields) is list:
                fields = ",".join(fields)
            if type(code) is list:
                code = ",".join(code)
            w.start()
            result = w.wss(code, fields, "tradeDate=" + trade_date + ";credibility=1").Data[0]
            if result[0] == u'CWSSService: invalid indicators.' and len(result) == 0:
                raise Exception("CWSSService: invalid indicators.")
            result = [0.0 if isnan(x) else x for x in result]
            return tuple(result) if len(result) > 1 else result[0]
        except Exception as e:
            if num_retries > 0:
                num_retries -= 1
                cls.wsi(code, fields, trade_date, num_retries=num_retries)

    @staticmethod
    def getMultiTimeSeriesDataFrame(codeList, beginDate, endDate, para, period="",
                                    tradingCalendar="", priceAdj="", credibility=0):
        """
        para只能是一个参数
        get time series from windPy, each code represents one capture
         月度合约: trade_hiscode
           :param credibility: (int)
           :param codeList: (list)
           :param beginDate: (date or datetime)
           :param endDate: (date or datetime)
           :param para: (string)只能是一个字符参数
           :param period: (int) 频率
           :param tradingCalendar: (string)   交易日历，选择可以选择银行间:NIB,不选择，则默认交易所日历
           :param priceAdj: (string) 价格是否调整,F:前复权，B:后复权
           :return: (DataFrame)
        """
        try:
            w.start()
            codeListStr = ",".join(codeList)
            period = ("Period=" + period) if period == "W" else ""
            tradingCalendar = ("TradingCalendar=" + tradingCalendar) if tradingCalendar != "" else ""
            priceAdj = ("priceAdj=" + priceAdj) if priceAdj != "" else ""
            credibility = ("credibility=" + str(credibility)) if credibility != 0 else ""
            windData = w.wsd(codeListStr,
                             para,
                             beginDate.strftime("%Y-%m-%d"),
                             endDate.strftime("%Y-%m-%d"),
                             period,
                             tradingCalendar,
                             priceAdj, credibility)
            if len(windData.Data) == 0:
                raise BaseException
            if len(windData.Data[0]) == 0:
                raise BaseException
            dataDict = {}
            for i in range(len(windData.Data)):
                dataDict[windData.Codes[i].lower() + "_" + para] = windData.Data[i]
            df = pd.DataFrame(dataDict, index=windData.Times)
            df.index = pd.to_datetime(df.index)
            df.index.name = "trade_date"
            return df
        except BaseException as e:
            print(format(e))
            raise

    @staticmethod
    def getTimeSeriesDataFrame(code, beginDate, endDate, paraList, period="",
                               tradingCalendar="", priceAdj="", credibility=0):
        """
        get time series from windPy, each code represents one capture
         月度合约: trade_hiscode
           :param credibility: (int)
           :param code: (string)
           :param beginDate: (date or datetime)
           :param endDate: (date or datetime)
           :param paraList: (list)
           :param period: (str) W or D 频率
           :param tradingCalendar: (string)   交易日历，选择可以选择银行间:NIB,不选择，则默认交易所日历
           :param priceAdj: (string) 价格是否调整,F:前复权，B:后复权
           :return: (DataFrame)
        """
        try:
            w.start()
            para = ",".join(paraList)
            period = ("Period=" + period) if period == "W" else ""
            tradingCalendar = ("TradingCalendar=" + tradingCalendar) if tradingCalendar != "" else ""
            priceAdj = ("priceAdj=" + priceAdj) if priceAdj != "" else ""
            credibility = ("credibility=" + str(credibility)) if credibility != 0 else ""
            windData = w.wsd(code,
                             para,
                             beginDate.strftime("%Y-%m-%d"),
                             endDate.strftime("%Y-%m-%d"),
                             tradingCalendar,
                             priceAdj, credibility)
            if len(windData.Data) == 0:
                raise BaseException
            if len(windData.Data[0]) == 0:
                raise BaseException
            dataDict = {}
            for i in range(len(windData.Data)):
                dataDict[windData.Fields[i].lower()] = windData.Data[i]
            df = pd.DataFrame(dataDict, index=windData.Times)
            df.index = pd.to_datetime(df.index)
            df.index.name = "trade_date"
            return df
        except BaseException as e:
            print(format(e))
            raise

    @staticmethod
    def getMinTimeSeriesDataFrame(code, beginDate, endDate, paraList, bar_size=1):
        """
        获取分钟级别数据
        get time series from windPy, each code represents one capture
         月度合约: trade_hiscode
           :param bar_size: (int)  The frequency of the data
           :param code: string
           :param beginDate: date or datetime
           :param endDate: date or datetime
           :param paraList: list
           :return: DataFrame
        """
        try:
            w.start()
            para = ",".join(paraList)
            bar_size = "" + str(bar_size) if bar_size is not None else ""
            windData = w.wsi(code,
                             para,
                             beginDate.strftime("%Y-%m-%d %H:%M:%S"),
                             endDate.strftime("%Y-%m-%d %H:%M:%S"), "")
            if len(windData.Data) == 0:
                raise BaseException
            if len(windData.Data[0]) == 0:
                raise BaseException
            dataDict = {}
            for i in range(len(windData.Data)):
                dataDict[windData.Fields[i].lower()] = windData.Data[i]
            df = pd.DataFrame(dataDict, index=windData.Times)
            if df.index[0].to_pydatetime().microsecond != 0:
                df.index -= timedelta(microseconds=df.index[0].to_pydatetime().microsecond)
            df.index.name = "trade_date"
            return df
        except BaseException as e:
            print(format(e))
            raise

    @staticmethod
    def getInfoDataFrame(code, paraList):
        """
        get info of one product by code
        :return: DataFrame
        :param code:
        :param paraList:
        :return:  DataFrame;
        """
        try:
            w.start()
            para = ",".join(paraList)
            windData = w.wss(code,
                             para)
            if len(windData.Data) == 0:
                return None
            if len(windData.Data[0]) == 0:
                return None
            dataDict = {}
            for i in range(len(windData.Data)):
                dataDict[windData.Fields[i].lower()] = windData.Data[i]
            df = pd.DataFrame(dataDict)
            df = df[paraList]
            return df
        except BaseException as e:
            print(format(e))
            raise

    @staticmethod
    def getEDBTimeSeriesDataFrame(codeList, beginDate, endDate, fillChoice="Previous"):
        """
        宏观数据提取
        get edb time series from windPy, each code represents one capture
        : Param fillChoice: (string) previous或者None，空值数据是否需要被前一日的数据取代
        """
        codeListStr = ",".join(codeList)
        try:
            w.start()
            if fillChoice == "Previous":
                windData = w.edb(codeListStr,
                                 beginDate.strftime("%Y-%m-%d"),
                                 endDate.strftime("%Y-%m-%d"),
                                 "Fill=" + fillChoice)
            else:
                windData = w.edb(codeListStr,
                                 beginDate.strftime("%Y-%m-%d"),
                                 endDate.strftime("%Y-%m-%d"))
            if len(windData.Data) == 0:
                return None
            if len(windData.Data[0]) == 0:
                return None
            dataDict = {}
            for i in range(len(windData.Data)):
                dataDict[windData.Codes[i]] = windData.Data[i]
            df = pd.DataFrame(dataDict, index=windData.Times)
            df.index = pd.to_datetime(df.index)
            df.index.name = "trade_date"
            return df
        except BaseException as e:
            print(format(e))
            raise

    @staticmethod
    def t_days_offset(offset=0, cur_date=datetime.now()):
        try:
            w.start()
            result = w.tdaysoffset(offset, cur_date, "").Data[0][0]
            # result = WindHelper.wsd(code="000001.SZ", paras=["close"], begin_date=datetime.now() - timedelta(days=10), end_date=datetime.now())
            # result = result.index[-1].to_pydatetime()
            return result
        except IndexError as e:
            print(format(e))
            raise

    @staticmethod
    def tdays_count(begin_date, end_date):
        w.start()
        result = w.tdayscount(begin_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d"), "").Data[0][0]
        return result

    @classmethod
    def all_tf_codes(cls, begin_date=None, end_date=None, contract_type="T"):
        """
        获取指定时间段内所有国债期货的合约
        :return type: list
        """
        if begin_date is None and contract_type == "T":
            begin_date = datetime(2015, 3, 20)
        elif begin_date is None and contract_type == "TF":
            begin_date = datetime(2013, 9, 6)
        elif begin_date is None and contract_type == "TS":
            begin_date = datetime(2018, 8, 17)
        if end_date is None:
            end_date = datetime.now()
        w.start()
        near_code = contract_type.upper() + "00.CFE"
        far_code = contract_type.upper() + "01.CFE"
        farfar_code = contract_type.upper() + "02.CFE"
        para = "trade_hiscode"
        near_df = cls.wsd(code=near_code, paras=para, begin_date=begin_date, end_date=end_date).dropna()
        # 国债期货下季列表
        far_df = cls.wsd(code=far_code, paras=para, begin_date=begin_date, end_date=end_date).dropna()
        # 国债期货隔季合约
        farfar_df = cls.wsd(code=farfar_code, paras=para, begin_date=begin_date, end_date=end_date).dropna()
        # 获取国债期货时间序列基础表：
        # 日期，当季合约，当季结算价，持仓量，下季合约，下季结算价，持仓量，隔季合约，隔季结算价，持仓量
        base_df = pd.DataFrame(near_df).append(far_df).append(farfar_df)
        if len(base_df) == 0:
            return None
        contract_code_list = base_df["trade_hiscode"].unique().tolist()
        return list(sorted(contract_code_list))

    @classmethod
    def tdays(cls, begin_date, end_date, tradingCalendar=""):
        """生成时间序列List"""
        w.start()
        return w.tdays(begin_date, end_date, tradingCalendar).Data[0]

    @classmethod
    def tf_dbs(cls, code, market="IB"):
        w.start()
        para = "windcode=" + code
        data = w.wset("conversionfactor", para).Data
        bondCodeList = data[0]
        cfList = data[1]
        dlvBdCodeDict = {}
        for i in range(len(bondCodeList)):
            bondCode = bondCodeList[i]
            if market in bondCode:
                dlvBdCodeDict[bondCode] = cfList[i]
        return dlvBdCodeDict

    @classmethod
    def get_end_date(cls):
        """确定开始时间，计算最新的已经结算的交易日"""
        # 确定结束时间
        # 结束时间为该合约的最后交易日和当前日期的最小值
        last_trade_date = cls.t_days_offset(offset=0, cur_date=datetime.now())
        # 确定结束时间
        if datetime.now().hour >= 19:  # 以晚上19点为界限
            end_date = last_trade_date
        elif datetime.now().date() > last_trade_date.date():  # 当天不是交易日
            end_date = last_trade_date
        else:  # 既非节假日，且当然的数据也没有生成
            end_date = cls.t_days_offset(offset=-1, cur_date=datetime.now())  # datetime类型
        return end_date

    @classmethod
    def get_dv01(cls, codes):
        """获取个券或者组合的dv01"""
        pass

    @classmethod
    def monthly_data_with_td(cls, df):
        """
        将月度宏观数据时间点和交易数据相对应
        :param df:
        :return:
        """
        begin_date = df.index[0]
        end_date = df.index[-1]
        items = cls.tdays(begin_date=begin_date, end_date=end_date)
        ds = [nearest(items, d) for d in df.index.to_pydatetime()]
        df.index = pd.DatetimeIndex(ds)
        return df

    @classmethod
    def monthly_data_with_signal(cls, df, shift):
        """
        将月度交易数据时间点和交易信号相对应
        :param shift:
        :param df:
        :return:
        """
        cls.monthly_data_with_td(df)
        return df.shift(shift)

    @classmethod
    def ffill(cls, df):
        return df.resample("M").last().ffill().dropna()

    @classmethod
    def fill(cls, df, end_date):
        df = df.resample("M").last().ffill().dropna()
        date_idx = pd.date_range(start=df.index[0], end=end_date)
        return pd.DataFrame(df, index=date_idx).ffill().dropna()



"""
#-#-#-#-#-#-#-#-#-#-#-#-#-#-季节性调整-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-

所有输入应当是月频时间序列，即以月频datetime为index的pandas.Series
需要下载的模组：numpy，pandas，scipy
关于季节性调整的资料参考以下网址：
http://cn.mathworks.com/help/econ/seasonal-adjustment-using-snxd7m-seasonal-filters.html

#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#--#-#-#-#-#-
"""

import numpy as np
import pandas as pd
import scipy.signal as sg


# -#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-滤波函数#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-

def mean_Jan_Feb(ts):
    """
    平均1、2月数据
    输入的时间序列中1月和2月数据必须成对出现！
    :param ts:
    :return:
    """
    ts_ave12 = ts.copy()
    Jan = pd.date_range(ts.index[0], ts.index[-1], freq='A-JAN')
    Feb = pd.date_range(ts.index[0], ts.index[-1], freq='A-FEB')
    ave12 = (ts.loc[Jan].values + ts.loc[Feb].values) / 2
    ts_ave12.loc[Jan] = ave12
    ts_ave12.loc[Feb] = ave12
    ts_ave12.name = 'mean_Jan_Feb'

    return ts_ave12


def MA13(ts):  # 13-Term Moving Average

    # Symmetric weights for 13-Term MA
    sw13 = np.zeros(13) + 1 / 12
    sw13[0] = 1 / 24
    sw13[-1] = 1 / 24

    arr_ts_MA13 = sg.convolve(ts, sw13, 'same')  # 卷积
    # the following two lines deals with data at end of Series
    arr_ts_MA13[0:6] = arr_ts_MA13[6]
    arr_ts_MA13[-6:] = arr_ts_MA13[-7]
    ts_MA13 = pd.Series(arr_ts_MA13, index=ts.index)
    ts_MA13.name = '13-Term Moving Average'

    return ts_MA13


def MA13C(ts):
    """
    13-Term Moving Average for centering data
    :param ts:
    :return:
    """
    # Symmetric weights for 13-Term MA
    sw13 = np.zeros(13) + 1 / 12
    sw13[0] = 1 / 24
    sw13[-1] = 1 / 24

    arr_ts_MA13C = sg.convolve(ts, sw13, 'same')
    # the following two lines deals with data at end of Series
    # different from original MA13
    arr_ts_MA13C[0:6] = arr_ts_MA13C[12:18]
    arr_ts_MA13C[-6:] = arr_ts_MA13C[-18:-12]
    ts_MA13C = pd.Series(arr_ts_MA13C, index=ts.index)
    ts_MA13C.name = '13-Term Moving Average C'

    return ts_MA13C


def H13(ts):  # 13-Term Henderson filter

    # Henderson filter weights
    swh = np.array([-0.019, -0.028, 0, .066, .147, .214,
                    .24, .214, .147, .066, 0, -0.028, -0.019])
    # Asymmetric weights for end of series
    awh = np.array([[-.034, -.017, .045, .148, .279, .421],
                    [-.005, .051, .130, .215, .292, .353],
                    [.061, .135, .201, .241, .254, .244],
                    [.144, .205, .230, .216, .174, .120],
                    [.211, .233, .208, .149, .080, .012],
                    [.238, .210, .144, .068, .002, -.058],
                    [.213, .146, .066, .003, -.039, -.092],
                    [.147, .066, .004, -.025, -.042, 0],
                    [.066, .003, -.020, -.016, 0, 0],
                    [.001, -.022, -.008, 0, 0, 0],
                    [-.026, -.011, 0, 0, 0, 0],
                    [-.016, 0, 0, 0, 0, 0]])

    # Apply 13-term Henderson filter
    arr_h13 = sg.convolve(ts, swh, 'same')
    arr_h13[-6:] = sg.convolve(ts[-12:].values[:, None], awh, 'valid')
    arr_h13[0:6] = sg.convolve(ts[0:12].values[:, None], np.rot90(awh, 2), 'valid')
    h13 = pd.Series(arr_h13, index=ts.index)
    h13.name = '13-Term Henderson Filter'

    return h13


def S33(ts):  # S3x3 seasonal components

    T = len(ts)
    sidx = []
    for i in range(12):
        sidx.append(np.arange(i, T, 12))

    # Symmetric weights
    sw3 = np.array([1 / 9, 2 / 9, 1 / 3, 2 / 9, 1 / 9])
    # Asymmetric weights for end of series
    aw3 = np.array([[.259, .407], [.37, .407], [.259, .185], [.111, 0]])
    aw3R = [[0, .111], [.185, .259], [.407, .37], [.407, .259]]
    # Apply filter to each month
    shat3 = ts.copy()
    for i in range(12):
        ns = len(sidx[i])
        dat = ts[sidx[i]].copy()
        sd = sg.convolve(dat.values, sw3, 'same')
        sd[0:2] = sg.convolve(dat[0:4].values[:, None], np.rot90(aw3, 2), 'valid')
        sd[-2:] = sg.convolve(dat[-4:].values[:, None], aw3, 'valid')
        shat3.iloc[sidx[i]] = sd

    # 13-term moving average of filtered series
    sb = MA13C(shat3)
    # Center to get final estimate
    s33 = shat3 / sb
    s33.name = 'Seasonal component estimates'

    return s33


def S35(ts):  # S3x5 seasonal Components

    T = len(ts)
    sidx = []
    for i in range(12):
        sidx.append(np.arange(i, T, 12))

    # Symmetric weights
    sw5 = np.zeros(7) + 1 / 5
    sw5[0:2] = [1 / 15, 2 / 15]
    sw5[-2:] = [2 / 15, 1 / 15]
    # Asymmetric weights for end of series
    aw5 = np.array([[.150, .250, .293],
                    [.217, .250, .283],
                    [.217, .250, .283],
                    [.217, .183, .150],
                    [.133, .067, 0],
                    [.067, 0, 0]])
    # Apply filter to each month
    shat5 = ts.copy()
    for i in range(12):
        ns = len(sidx[i])
        dat = ts[sidx[i]].copy()
        sd = sg.convolve(dat.values, sw5, 'same')
        sd[0:3] = sg.convolve(dat[0:6].values[:, None], np.rot90(aw5, 2), 'valid')
        sd[-3:] = sg.convolve(dat[-6:].values[:, None], aw5, 'valid')
        shat5.iloc[sidx[i]] = sd

    # 13-term moving average of filtered series
    sb = MA13C(shat5)
    # Center to get final estimate
    s35 = shat5 / sb
    s35.name = 'Seasonal component estimates'

    return s35


# -#-#-#-#-#-#-#-derive deseasonalized series-#-#-#-#-#-#-#-

def Deseason(ts, Jan_Feb=False):
    """
    # -----是否平均1、2月数据-----
    :param ts:
    :param Jan_Feb:
    :return:
    """
    if Jan_Feb:  # 如果选择平均，则对1,2月数据进行取平均，否则就不做处理
        IVA_av12 = mean_Jan_Feb(ts)
    else:
        IVA_av12 = ts

    # -----Detrend the data using a 13-term moving average-----
    IVA_MA13 = MA13(IVA_av12)
    IVA_xt = IVA_av12 / IVA_MA13

    # -----Apply an S(3,3) seasonal filter to deseasonalize-----
    IVA_S33 = S33(IVA_xt)
    IVA_DeSeason = IVA_av12 / IVA_S33  # Deseasonalize series

    # -----Apply a 13-term Henderson filter to improve detrending-----
    IVA_H13 = H13(IVA_DeSeason)
    IVA_xt2 = IVA_av12 / IVA_H13  # New detrended series

    # -----Apply an S(3,5) seasonal filter to improve deseasonalizing-----
    IVA_S35 = S35(IVA_xt2)
    IVA_DeSeason2 = IVA_av12 / IVA_S35
    IVA_DeSeason2.name = 'Deseasonalized IVA'

    return IVA_DeSeason2

################### X13 ########################



from datetime import timedelta


MONTH_NAME = ["", "January", "February", "March", "April", "May", "June", "July", "August", "September", "October",
              "November", "December"]
MONTH_DAYS = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

LUNAR_CALENDAR_TABLE = [
    0x04AE53, 0x0A5748, 0x5526BD, 0x0D2650, 0x0D9544, 0x46AAB9, 0x056A4D, 0x09AD42, 0x24AEB6, 0x04AE4A,
    # //*1901-1910*/
    0x6A4DBE, 0x0A4D52, 0x0D2546, 0x5D52BA, 0x0B544E, 0x0D6A43, 0x296D37, 0x095B4B, 0x749BC1, 0x049754,
    # //*1911-1920*/
    0x0A4B48, 0x5B25BC, 0x06A550, 0x06D445, 0x4ADAB8, 0x02B64D, 0x095742, 0x2497B7, 0x04974A, 0x664B3E,
    # //*1921-1930*/
    0x0D4A51, 0x0EA546, 0x56D4BA, 0x05AD4E, 0x02B644, 0x393738, 0x092E4B, 0x7C96BF, 0x0C9553, 0x0D4A48,
    # //*1931-1940*/
    0x6DA53B, 0x0B554F, 0x056A45, 0x4AADB9, 0x025D4D, 0x092D42, 0x2C95B6, 0x0A954A, 0x7B4ABD, 0x06CA51,
    # //*1941-1950*/
    0x0B5546, 0x555ABB, 0x04DA4E, 0x0A5B43, 0x352BB8, 0x052B4C, 0x8A953F, 0x0E9552, 0x06AA48, 0x6AD53C,
    # //*1951-1960*/
    0x0AB54F, 0x04B645, 0x4A5739, 0x0A574D, 0x052642, 0x3E9335, 0x0D9549, 0x75AABE, 0x056A51, 0x096D46,
    # //*1961-1970*/
    0x54AEBB, 0x04AD4F, 0x0A4D43, 0x4D26B7, 0x0D254B, 0x8D52BF, 0x0B5452, 0x0B6A47, 0x696D3C, 0x095B50,
    # //*1971-1980*/
    0x049B45, 0x4A4BB9, 0x0A4B4D, 0xAB25C2, 0x06A554, 0x06D449, 0x6ADA3D, 0x0AB651, 0x093746, 0x5497BB,
    # //*1981-1990*/
    0x04974F, 0x064B44, 0x36A537, 0x0EA54A, 0x86B2BF, 0x05AC53, 0x0AB647, 0x5936BC, 0x092E50, 0x0C9645,
    # //*1991-2000*/
    0x4D4AB8, 0x0D4A4C, 0x0DA541, 0x25AAB6, 0x056A49, 0x7AADBD, 0x025D52, 0x092D47, 0x5C95BA, 0x0A954E,
    # //*2001-2010*/
    0x0B4A43, 0x4B5537, 0x0AD54A, 0x955ABF, 0x04BA53, 0x0A5B48, 0x652BBC, 0x052B50, 0x0A9345, 0x474AB9,
    # //*2011-2020*/
    0x06AA4C, 0x0AD541, 0x24DAB6, 0x04B64A, 0x69573D, 0x0A4E51, 0x0D2646, 0x5E933A, 0x0D534D, 0x05AA43,
    # //*2021-2030*/
    0x36B537, 0x096D4B, 0xB4AEBF, 0x04AD53, 0x0A4D48, 0x6D25BC, 0x0D254F, 0x0D5244, 0x5DAA38, 0x0B5A4C,
    # //*2031-2040*/
    0x056D41, 0x24ADB6, 0x049B4A, 0x7A4BBE, 0x0A4B51, 0x0AA546, 0x5B52BA, 0x06D24E, 0x0ADA42, 0x355B37,
    # //*2041-2050*/
    0x09374B, 0x8497C1, 0x049753, 0x064B48, 0x66A53C, 0x0EA54F, 0x06B244, 0x4AB638, 0x0AAE4C, 0x092E42,
    # //*2051-2060*/
    0x3C9735, 0x0C9649, 0x7D4ABD, 0x0D4A51, 0x0DA545, 0x55AABA, 0x056A4E, 0x0A6D43, 0x452EB7, 0x052D4B,
    # //*2061-2070*/
    0x8A95BF, 0x0A9553, 0x0B4A47, 0x6B553B, 0x0AD54F, 0x055A45, 0x4A5D38, 0x0A5B4C, 0x052B42, 0x3A93B6,
    # //*2071-2080*/
    0x069349, 0x7729BD, 0x06AA51, 0x0AD546, 0x54DABA, 0x04B64E, 0x0A5743, 0x452738, 0x0D264A, 0x8E933E,
    # //*2081-2090*/
    0x0D5252, 0x0DAA47, 0x66B53B, 0x056D4F, 0x04AE45, 0x4A4EB9, 0x0A4D4C, 0x0D1541, 0x2D92B5  # //*2091-2099*/
]


# 下面的三个表格是农历数据表 LunarCalendarTable 的结构。总共使用了32位整数的0～23位。
#
# 6 5                4 3 2 1 0
# 表示春节的公历月份 表示春节的公历日期
#
# 19 18 17 16 15 14 13 12 11 10  9  8  7
# 1   2  3  4  5  6  7  8  9 10 11 12 13
# 农历1-13 月大小 。月份对应位为1，农历月大(30 天),为0 表示小(29 天)
#
# 23 22 21 20
# 表示当年闰月月份，值为0 为则表示当年无闰月。

def get_month_days(year, month):
    global MONTH_DAYS
    if month == 2:
        if ((year % 4 == 0) and (year % 100 != 0)) or (year % 400 == 0):
            return 29
        else:
            return 28
    else:
        return MONTH_DAYS[month];


def get_syear_days(syear):
    if ((syear % 4 == 0) and (syear % 100 != 0)) or (syear % 400 == 0):
        return 366
    else:
        return 365


def get_days_of_syear(syear, smonth, sday):
    """ get given day's number of sun year """
    days = 0
    for i in range(1, smonth):
        days += get_month_days(syear, i)
    days += sday
    return days


def get_days_of_lyear(syear, smonth, sday):
    """ get given day's number of the lunar year """
    global LUNAR_CALENDAR_TABLE
    lyear = syear
    spring_month = (LUNAR_CALENDAR_TABLE[syear - 1901] & 0x60) >> 5
    spring_day = (LUNAR_CALENDAR_TABLE[syear - 1901] & 0x1F)
    if (spring_month > smonth) or ((spring_month == smonth) and (spring_day > sday)):
        # the day is before spring festival day, and is previous day in lunar year
        spring_month = (LUNAR_CALENDAR_TABLE[syear - 1901 - 1] & 0x60) >> 5
        spring_day = (LUNAR_CALENDAR_TABLE[syear - 1901 - 1] & 0x1F)
        lyear -= 1
        lunar_days = get_syear_days(lyear) + get_days_of_syear(syear, smonth, sday) \
                     - get_days_of_syear(lyear, spring_month, spring_day)
    else:
        lunar_days = get_days_of_syear(syear, smonth, sday) \
                     - get_days_of_syear(syear, spring_month, spring_day)
    lunar_days += 1  # consider current day
    return lyear, lunar_days


def get_lunar_date(syear, smonth, sday):
    if syear < 1901 or syear > 2099:
        return
        # lunar year, lunar days to spring festival
    lyear, lunar_days = get_days_of_lyear(syear, smonth, sday)
    l_double_month = (LUNAR_CALENDAR_TABLE[lyear - 1901] >> 20) & 0xF

    lmonth = lday = 1
    bits = 19
    month_begin_day = 0
    for lmonth in range(1, 14):
        l_month_big = (LUNAR_CALENDAR_TABLE[lyear - 1901] >> bits) & 0x1
        if month_begin_day + 29 + l_month_big < lunar_days:
            lmonth += 1
            month_begin_day += 29 + l_month_big
        else:
            lday = lunar_days - month_begin_day
            break
        bits -= 1
    if l_double_month:
        # lunar double month adjust
        if l_double_month == lmonth - 1:
            lmonth -= 1
            lmonth += 100  # double month
        elif l_double_month < lmonth - 1:
            lmonth -= 1
    return lyear, lmonth, lday


# ******************************************************************************
# 下面为阴历计算所需的数据,为节省存储空间,所以采用下面比较变态的存储方法.
# ******************************************************************************
# 数组g_lunar_month_day存入阴历1901年到2050年每年中的月天数信息，
# 阴历每月只能是29或30天，一年用12（或13）个二进制位表示，对应位为1表30天，否则为29天
g_lunar_month_day = [
    0x4ae0, 0xa570, 0x5268, 0xd260, 0xd950, 0x6aa8, 0x56a0, 0x9ad0, 0x4ae8, 0x4ae0,  # 1910
    0xa4d8, 0xa4d0, 0xd250, 0xd548, 0xb550, 0x56a0, 0x96d0, 0x95b0, 0x49b8, 0x49b0,  # 1920
    0xa4b0, 0xb258, 0x6a50, 0x6d40, 0xada8, 0x2b60, 0x9570, 0x4978, 0x4970, 0x64b0,  # 1930
    0xd4a0, 0xea50, 0x6d48, 0x5ad0, 0x2b60, 0x9370, 0x92e0, 0xc968, 0xc950, 0xd4a0,  # 1940
    0xda50, 0xb550, 0x56a0, 0xaad8, 0x25d0, 0x92d0, 0xc958, 0xa950, 0xb4a8, 0x6ca0,  # 1950
    0xb550, 0x55a8, 0x4da0, 0xa5b0, 0x52b8, 0x52b0, 0xa950, 0xe950, 0x6aa0, 0xad50,  # 1960
    0xab50, 0x4b60, 0xa570, 0xa570, 0x5260, 0xe930, 0xd950, 0x5aa8, 0x56a0, 0x96d0,  # 1970
    0x4ae8, 0x4ad0, 0xa4d0, 0xd268, 0xd250, 0xd528, 0xb540, 0xb6a0, 0x96d0, 0x95b0,  # 1980
    0x49b0, 0xa4b8, 0xa4b0, 0xb258, 0x6a50, 0x6d40, 0xada0, 0xab60, 0x9370, 0x4978,  # 1990
    0x4970, 0x64b0, 0x6a50, 0xea50, 0x6b28, 0x5ac0, 0xab60, 0x9368, 0x92e0, 0xc960,  # 2000
    0xd4a8, 0xd4a0, 0xda50, 0x5aa8, 0x56a0, 0xaad8, 0x25d0, 0x92d0, 0xc958, 0xa950,  # 2010
    0xb4a0, 0xb550, 0xb550, 0x55a8, 0x4ba0, 0xa5b0, 0x52b8, 0x52b0, 0xa930, 0x74a8,  # 2020
    0x6aa0, 0xad50, 0x4da8, 0x4b60, 0x9570, 0xa4e0, 0xd260, 0xe930, 0xd530, 0x5aa0,  # 2030
    0x6b50, 0x96d0, 0x4ae8, 0x4ad0, 0xa4d0, 0xd258, 0xd250, 0xd520, 0xdaa0, 0xb5a0,  # 2040
    0x56d0, 0x4ad8, 0x49b0, 0xa4b8, 0xa4b0, 0xaa50, 0xb528, 0x6d20, 0xada0, 0x55b0,  # 2050
]

# 数组gLanarMonth存放阴历1901年到2050年闰月的月份，如没有则为0，每字节存两年
g_lunar_month = [
    0x00, 0x50, 0x04, 0x00, 0x20,  # 1910
    0x60, 0x05, 0x00, 0x20, 0x70,  # 1920
    0x05, 0x00, 0x40, 0x02, 0x06,  # 1930
    0x00, 0x50, 0x03, 0x07, 0x00,  # 1940
    0x60, 0x04, 0x00, 0x20, 0x70,  # 1950
    0x05, 0x00, 0x30, 0x80, 0x06,  # 1960
    0x00, 0x40, 0x03, 0x07, 0x00,  # 1970
    0x50, 0x04, 0x08, 0x00, 0x60,  # 1980
    0x04, 0x0a, 0x00, 0x60, 0x05,  # 1990
    0x00, 0x30, 0x80, 0x05, 0x00,  # 2000
    0x40, 0x02, 0x07, 0x00, 0x50,  # 2010
    0x04, 0x09, 0x00, 0x60, 0x04,  # 2020
    0x00, 0x20, 0x60, 0x05, 0x00,  # 2030
    0x30, 0xb0, 0x06, 0x00, 0x50,  # 2040
    0x02, 0x07, 0x00, 0x50, 0x03  # 2050
]

# ==================================================================================

from datetime import date, datetime
from calendar import Calendar as Cal

START_YEAR = 1901


def is_leap_year(tm):
    y = tm.year
    return (not (y % 4)) and (y % 100) or (not (y % 400))


def show_month(tm):
    """

    :param tm: 输入年月日
    :return:
    """
    (ly, lm, ld) = get_ludar_date(tm)
    print()
    print("%d年%d月%d日" % (tm.year, tm.month, tm.day), week_str(tm),)
    print("\t农历：", y_lunar(ly), m_lunar(lm), d_lunar(ld))
    print("日\t一\t二\t三\t四\t五\t六")

    c = Cal()
    ds = [d for d in c.itermonthdays(tm.year, tm.month)]
    count = 0
    for d in ds:
        count += 1
        if d == 0:
            print("\t")
            continue

        (ly, lm, ld) = get_ludar_date(datetime(tm.year, tm.month, d))
        if count % 7 == 0:
            print()

        d_str = str(d)
        if d == tm.day:
            d_str = u"*" + d_str
        print("")
        print(d_str + d_lunar(ld) + u"\t")
    print()


def this_month():
    show_month(datetime.now())


def week_str(tm):
    a = '星期一 星期二 星期三 星期四 星期五 星期六 星期日'.split()
    return a[tm.weekday()]


def d_lunar(ld):
    a = '初一 初二 初三 初四 初五 初六 初七 初八 初九 初十\
    十一 十二 十三 十四 十五 十六 十七 十八 十九 廿十 \
            廿一 廿二 廿三 廿四 廿五 廿六 廿七 廿八 廿九 三十'.split()
    return a[ld - 1]


def m_lunar(lm):
    a = '正月 二月 三月 四月 五月 六月 七月 八月 九月 十月 十一月 十二月'.split()
    return a[lm - 1]


def y_lunar(ly):
    y = ly
    tg = '甲 乙 丙 丁 戊 己 庚 辛 壬 癸'.split()
    dz = '子 丑 寅 卯 辰 巳 午 未 申 酉 戌 亥'.split()
    sx = '鼠 牛 虎 免 龙 蛇 马 羊 猴 鸡 狗 猪'.split()
    return tg[(y - 4) % 10] + dz[(y - 4) % 12] + u' ' + sx[(y - 4) % 12] + u'年'


def date_diff(tm):
    return (tm - datetime(1901, 1, 1)).days


def get_leap_month(lunar_year):
    flag = g_lunar_month[int((lunar_year - START_YEAR) / 2)]
    if (lunar_year - START_YEAR) % 2:
        return flag & 0x0f
    else:
        return flag >> 4


def lunar_month_days(lunar_year, lunar_month):
    if lunar_year < START_YEAR:
        return 30

    high, low = 0, 29
    iBit = 16 - lunar_month

    if lunar_month > get_leap_month(lunar_year) and get_leap_month(lunar_year):
        iBit -= 1

    if g_lunar_month_day[lunar_year - START_YEAR] & (1 << iBit):
        low += 1

    if lunar_month == get_leap_month(lunar_year):
        if g_lunar_month_day[lunar_year - START_YEAR] & (1 << (iBit - 1)):
            high = 30
        else:
            high = 29

    return high, low


def lunar_year_days(year):
    days = 0
    for i in range(1, 13):
        (high, low) = lunar_month_days(year, i)
        days += high
        days += low
    return days


def get_ludar_date(tm):
    """
    输入日期，返回农历日期的年，月，日
    :param tm:
    :return:
    """
    span_days = date_diff(tm)

    # 阳历1901年2月19日为阴历1901年正月初一
    # 阳历1901年1月1日到2月19日共有49天
    if span_days < 49:
        year = START_YEAR - 1
        if span_days < 19:
            month = 11
            day = 11 + span_days
        else:
            month = 12
            day = span_days - 18
        return year, month, day

        # 下面从阴历1901年正月初一算起
    span_days -= 49
    year, month, day = START_YEAR, 1, 1
    # 计算年
    tmp = lunar_year_days(year)
    while span_days >= tmp:
        span_days -= tmp
        year += 1
        tmp = lunar_year_days(year)

        # 计算月
    (foo, tmp) = lunar_month_days(year, month)
    while span_days >= tmp:
        span_days -= tmp
        if month == get_leap_month(year):
            (tmp, foo) = lunar_month_days(year, month)
            if span_days < tmp:
                return 0, 0, 0
            span_days -= tmp
        month += 1
        (foo, tmp) = lunar_month_days(year, month)

        # 计算日
    day += span_days
    return year, month, day


def get_lunar_datetime(cur_date):
    """输入阳历日期，返回农历日期"""
    year, month, day = get_lunar_date(cur_date)
    return datetime(year, month, day)


def spring_month(year):
    """
    输入年份,返回春节所在的月份
    :return:
    """
    jan = datetime(year, 1, 1)
    feb = datetime(year, 2, 1)
    mar = datetime(year, 3, 1)
    apr = datetime(year, 4, 1)
    jan_end = feb + timedelta(days=-1)
    feb_end = mar + timedelta(days=-1)
    mar_end = apr + timedelta(days=-1)
    jan_m, jan_d = get_ludar_date(jan)[1:]
    feb_m, feb_d = get_ludar_date(feb)[1:]
    mar_m, mar_d = get_ludar_date(mar)[1:]
    apr_m, apr_d = get_ludar_date(apr)[1:]
    # jan_end_m, jan_end_d = get_ludar_date(jan)[1:]
    # feb_end_m, feb_end_d = get_ludar_date(feb)[1:]
    # mar_end_m, mar_end_d = get_ludar_date(mar)[1:]
    if jan_m == 12:
        if feb_d > 1:
            return jan.month
        else:
            return feb.month
    elif feb_m == 12:
        if mar_d > 1:
            return feb.month
        else:
            return mar.month
    elif mar_m == 12:
        if apr_d > 1:
            return mar.month
        else:
            return apr.month
    # 如果没有出现12
    elif jan_m == 11:
        if feb_d > 1:
            return jan.month
        else:
            return feb.month
    elif feb_m == 11:
        if mar_d > 1:
            return feb.month
        else:
            return mar.month


if __name__ == "__main__":
    y, m, d = 2018, 9, 13
    print("Sun calendar 2010-9-13 == Lunar calendar ", get_lunar_date(y, m, d))
    # this_month()
    spring_month(2003)
    for i in range(1990, 2020):
        print(str(i) + ":" + str(spring_month(i)))



import pandas as pd
import statsmodels as sm
from pyfi.wind_helper import *
from pyfi.common import get_end_date
import os
import matplotlib.pyplot as plt
from statsmodels.tsa.x13 import *
from datetime import datetime

begin_date = datetime(2005, 1, 1)
end_date = get_end_date()
df = ds_ip_idx(begin_date, end_date).dropna()
df.name = "ip"

XPATH = os.chdir("C:\\dev_env\\WinX13\\x13as")

temp_series = pd.DataFrame(df.values, index=df.index)
temp_series.index.fre_str = "M"
x13results = x13_arima_analysis(endog=temp_series, x12path=XPATH, outlier=True, print_stdout=True)

df_adj = pd.Series(x13results.seasadj.values, index=df.index.values)
df_adj = df_adj.to_frame()
df_adj.columns = ["Adjusted"]
os.chdir(os.getcwd())

ip = WindHelper.edb(codes=["ip_yoy"], begin_date=begin_date, end_date=end_date)/100.0

fig, ax = plt.subplots(figsize=(13, 6))
ax.set_title("季调数据")
ax.plot(df_adj.index, df_adj["Adjusted"], linewidth=2, marker='', markersize=3, zorder=1, label="Seasonally Adjusted")
ax.plot(df_adj.index, ip["ip_yoy"], linewidth=2, marker='', markersize=3, zorder=1, label="Actual")
plt.show()



maxdd = lambda eqd: (eqd.cumsum().expanding().max() - eqd.cumsum()).max()


def m_sharpe(eqd, bk=0.0):
    """monthly sharpe ratio"""
    try:
        if eqd.sum() == 0:
            return -100
        return ((eqd.mean() - bk / 12) / eqd.std()) * (12 ** 0.5)
    except BaseException as e:
        print(e)


def excess_msp(eqd, base_eqd, bk=0.0):
    return m_sharpe(eqd, bk) - m_sharpe(base_eqd, bk)


import matplotlib.pyplot as plt
import matplotlib as mpl


def init():
    """
    一些初始化设置，主要是为了显示负号和中文
    :return:
    """
    mpl.rcParams["axes.unicode_minus"] = False
    plt.rcParams["font.sans-serif"] = ["SimHei"]
    plt.rcParams["axes.unicode_minus"] = False
    from pandas.plotting import register_matplotlib_converters
    register_matplotlib_converters()


def line_graph(series, title="multi lines graph", legend_list="", legend_config=(20, 0), fig=None, ax=None, save=False):
    """df_list的时间轴得一样
    :param legend_config:
    :param save: 保存图片
    :param series:
    :param title:
    :param legend_list:
    :param fig: 
    :param ax:
    """
    init()
    if fig is None:
        fig = plt.figure(figsize=(16, 6))
    if ax is None:
        ax = plt.subplot(111)
    ax.set_title(title, fontsize=14)
    ax.grid(axis="both", linestyle='--')
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(True)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(True)
    ax.tick_params(axis='both', labelsize=16)
    # ax_list = [{}] * len(series_list)
    for i in range(len(series)):
        ax.plot(series[i].index.values, series[i].values, lw=2., linestyle="-")
        for tick in ax.get_xticklabels():
            tick.set_rotation(30)
    ax.legend(legend_list, fontsize=legend_config[0], loc=legend_config[1])
    if save:
        fig.savefig(title + ".jpg")


def double_lines(series1,
                 series2,
                 lgd1="series1",
                 lgd2="series2",
                 title="double axis graph",
                 colors=("grey", "red"),
                 figname="",
                 fig=None,
                 ax=None):
    """快速画出双线对比图
    :param series1:
    :param series2:
    :param lgd1:
    :param lgd2:
    :param title:
    :param colors:
    :param figname:
    :param fig:
    :param ax:
    """
    init()
    if fig is None:
        fig = plt.figure(figsize=(16, 9))
    if ax is None:
        ax1 = plt.subplot(111)
    else:
        ax1 = ax
    ax1.set_title(title, fontsize=18)
    ax1.grid(axis="both", linestyle='--')
    ax1.spines["top"].set_visible(False)
    ax1.spines["bottom"].set_visible(True)
    ax1.spines["right"].set_visible(False)
    ax1.spines["left"].set_visible(True)
    ax1.tick_params(axis='both', labelsize=16)
    if type(series1) is not list:
        ax1.plot(series1.index.values, series1.values, color=colors[0])
        ax1.legend([lgd1], fontsize=16, loc=2)
    else:
        for i in range(len(series1)):
            ax1.plot(series1[i].index.values, series1[i].values, lw=2., linestyle="-")
            ax1.legend(lgd1, fontsize=16, loc=2)
    ax2 = ax1.twinx()
    if type(series2) is not list:
        ax2.plot(series2.index.values, series2.values, color=colors[1])
        ax2.legend([lgd2], fontsize=16, loc=1)
    else:
        for i in range(len(series2)):
            ax2.plot(series2[i].index.values, series2[i].values, lw=2., linestyle="-")
            ax2.legend(lgd2, fontsize=16, loc=1)
    ax2.tick_params(axis='both', labelsize=16)
    if figname is not "":
        fig.savefig(figname)
    # plt.show()


def radar2_graph():
    import numpy as np
    import pylab as pl

    class Radar(object):

        def __init__(self, fig, titles, labels, rect=None):
            if rect is None:
                rect = [0.05, 0.05, 0.95, 0.95]
            self.n = len(titles)
            self.angles = np.arange(90, 90 + 360, 360.0 / self.n)
            self.axes = [fig.add_axes(rect, projection="polar", label="axes%d" % i)
                         for i in range(self.n)]

            self.ax = self.axes[0]
            self.ax.set_thetagrids(self.angles, labels=titles, fontsize=14)

            for ax in self.axes[1:]:
                ax.patch.set_visible(False)
                ax.grid("off")
                ax.xaxis.set_visible(False)

            for ax, angle, label in zip(self.axes, self.angles, labels):
                ax.set_rgrids(range(1, 6), angle=angle, labels=label)
                ax.spines["polar"].set_visible(False)
                ax.set_ylim(0, 5)

        def plot(self, values, *args, **kw):
            angle = np.deg2rad(np.r_[self.angles, self.angles[0]])
            values = np.r_[values, values[0]]
            self.ax.plot(angle, values, *args, **kw)

    fig = pl.figure(figsize=(6, 6))

    titles = list("ABCDE")

    labels = [
        list("abcde"), list("12345"), list("uvwxy"),
        ["one", "two", "three", "four", "five"],
        list("jklmn")
    ]

    radar = Radar(fig, titles, labels)
    radar.plot([1, 3, 2, 5, 4], "-", lw=2, color="b", alpha=0.4, label="first")
    radar.plot([2.3, 2, 3, 3, 2], "-", lw=2, color="r", alpha=0.4, label="second")
    radar.plot([3, 4, 3, 4, 2], "-", lw=2, color="g", alpha=0.4, label="third")
    radar.ax.legend()


def KDEplot(data, start, end, kde_bin, hist_bin):
    from scipy import stats
    import numpy as np
    ind = np.linspace(start, end, kde_bin)
    gkde = stats.gaussian_kde(data)
    kdepdf = gkde.evaluate(ind)
    plt.style.use('seaborn')
    fig, left_axis = plt.subplots(figsize=(20, 10))
    right_axis = left_axis.twinx()
    left_axis.hist(data, bins=hist_bin, label='hist')
    left_axis.grid(False)
    left_axis.legend(loc=2, prop={'size': 20})
    left_axis.tick_params(labelsize=15)
    left_axis.set_ylabel('historical frequency', fontsize=15)
    left_axis.grid(False)
    right_axis.plot(ind, kdepdf, label='kde', color="r")
    right_axis.grid(False)
    right_axis.legend(loc=1, prop={'size': 20})
    right_axis.tick_params(labelsize=15)
    right_axis.set_ylabel('kde probability', fontsize=15)


def multi_lines(data, cols=None, spacing=.1, figname="", **kwargs):
    from pandas import plotting
    import matplotlib.pyplot as plt
    init()
    fig = plt.figure(figsize=(16, 9))
    if cols is None:
        cols = data.columns
    if len(cols) == 0:
        return
    # Get default color style from pandas - can be changed to any other color list
    colors = getattr(getattr(plotting, '_style'), '_get_standard_colors')(num_colors=len(cols))

    # First axis
    ax = plt.subplot(111)
    ax = data.loc[:, cols[0]].plot(label=cols[0], color=colors[0], ax=ax, **kwargs)
    ax.set_ylabel(ylabel=cols[0])
    lines, labels = ax.get_legend_handles_labels()

    for n in range(1, len(cols)):
        # Multiple y-axes
        plt.grid(False)
        ax_new = ax.twinx()
        ax_new.spines['right'].set_position(('axes', 1 + spacing * (n - 1)))
        data.loc[:, cols[n]].plot(ax=ax_new, label=cols[n], color=colors[n % len(colors)])
        ax_new.set_ylabel(ylabel=cols[n])
        # Proper legend position
        line, label = ax_new.get_legend_handles_labels()
        lines += line
        labels += label
    ax.legend(lines, labels, loc=0)
    if figname is not "":
        fig.savefig(figname)
    # plt.show()




# 常用债券计算函数
from datetime import timedelta, datetime
import scipy.optimize as optimize
import pandas as pd
import numpy as np


def get_ytm_by_net(net, maturity, cupn, freq, par=100, guess=3):
    """
    根据净价, 待偿期限，票息，付息次数，面值计算债券的到期收益率
    :param net:
    :param par:
    :param maturity:
    :param cupn:
    :param freq:
    :param guess:
    :return:
    """
    yearsToNxCupn = maturity - 1. / freq * np.floor(freq * maturity)
    cupnCount = int(np.ceil(freq * maturity))
    coupon = cupn / 100. * par / freq
    if yearsToNxCupn == 0.:
        dt = [(i + 1.) / freq for i in range(cupnCount)]
    else:
        dt = [yearsToNxCupn + i / freq for i in range(cupnCount)]
    ytm_func = lambda y: sum([coupon / (1. + y / 100. / freq) ** (freq * t) for t in dt]) + \
                         par / (1. + y / 100. / freq) ** (freq * maturity) - (
                                 1. / freq - yearsToNxCupn) * cupn - net
    return optimize.newton(ytm_func, guess)


def get_dirty(ytm, maturity, cupn, freq, par=100.):
    """
    based on ytm, maturity, coupon and frequency to calculate the dirty price
    :param ytm:
    :param par:
    :param maturity:
    :param cupn: 3.65 instead of 0.0365
    :param freq:
    :return:
    """
    yearsToNxCupn = maturity - 1.0 / freq * np.floor(freq * maturity)
    cupnCount = int(np.floor(freq * maturity)) + 1
    coupon = cupn / 100. * par / freq
    if yearsToNxCupn == 0:
        dt = [(i + 1.) / freq for i in range(cupnCount)]  # 记录付息时间点
    else:
        dt = [yearsToNxCupn + i / freq for i in range(cupnCount)] # 记录付息时间点
    return sum([coupon / (1. + ytm / 100. / freq) ** (freq * t) for t in dt]) + \
        par / (1. + ytm / 100. / freq) ** (freq * maturity)


def get_net(ytm, par, maturity, cupn, freq):
    daysToNxCupn = maturity - 1.0 / freq * np.floor(freq * maturity)
    dirty = get_dirty(ytm, maturity, cupn, freq, par)
    return dirty - (1. / freq - daysToNxCupn) * cupn


def get_years_to_nxcupn(maturity, freq):
    """
    calculate the period to next coupon marked by year unit
    :param maturity:
    :param freq:
    :return:
    """
    yearsToNxCupn = maturity - 1.0 / freq * np.floor(freq * maturity)
    return yearsToNxCupn


import pandas as pd
from datetime import datetime

"""
本部分代码主要用于固化一些常规的指标，并用一定的命名规范对数据进行预处理：
细则如下：
- eco开头表示经济层面
- inf开头表示通胀层面
- val开头表示估值层面
- mon开头表示资金层面
- cyc开头表示周期层面
- sea开头表示海外层面
- tec开头表示技术情绪层面
"""


def gz10y(begin_date, end_date):
    """
    社融
    :param begin_date:
    :param end_date:
    :return:
    """
    from pyfi import WindHelper
    ytm = WindHelper.edb(codes=["gz10y"],
                         begin_date=begin_date,
                         end_date=end_date).iloc[:, 0]
    return ytm


def eco_sf(begin_date="2004-01-31", end_date="2018-01-31"):
    """
    社融存量同比
    2016年以前的数据，存量年度增量和月度增量和的差按照加权比例分配到每个月进行核销
    :return:
    """
    from pyfi import WindHelper
    if begin_date < datetime(2004, 1, 31):
        raise Exception("社融起始时间太早啦")
    init_reserve = 148532.0  # 2002年末
    inc = WindHelper.edb(codes=["M5206730"],
                         begin_date=datetime(2003, 1, 1),
                         end_date=end_date, adjust=False)
    reserve = inc.cumsum() + init_reserve
    return (reserve / reserve.shift(12) - 1).iloc[:, 0].dropna()


def eco_ip(begin_date="2002-02-01", end_date="2018-01-31", method=2, adjust=False):
    """
    获取预处理之后的工业增加值数据
    :param method: 1: 删除 or 2: 沿用12月数据
    :param adjust: 是否交易日调整
    :param begin_date:
    :param end_date:
    :return:
    """

    from pyfi import WindHelper, macro_adjust
    start_date = "2002-02-01" if type(begin_date) is str else datetime(2002, 2, 1)
    ip = WindHelper.edb(codes=["ip_yoy", "ip_cyoy"],
                        begin_date=start_date,
                        end_date=end_date, adjust=adjust)
    ip = macro_adjust(ip.ip_yoy, ip.ip_cyoy, method=method)  # 2月份替换为累计同比
    return ip.loc[(ip.index >= begin_date) & (ip.index <= end_date)]


def eff_rmb_idx(begin_date="2010-01-01", end_date="2018-01-02"):
    """
    生成CFETS人民币汇率指数，并用BIS人民币实际汇率指数作为前期补充，该数据为月度数据
    :param begin_date:
    :param end_date:
    :return:
    """
    from pyfi import WindHelper as W
    mid_date = "2015-11-30" if type(end_date) is str else datetime(2015, 11, 30)
    bis = W.edb(["M0000209"], begin_date, mid_date).resample("M").mean().ffill().iloc[:, 0]
    cfets = W.edb(["M0325601"], mid_date, end_date).resample("M").mean().ffill().iloc[:, 0]
    data = ((bis.iloc[:-1]) * cfets[0] / bis[-1]).append(cfets)
    return data


def ds_ip_idx(begin_date, end_date):
    """
    去除季节性之后的ip的定基指数
    :param begin_date:
    :param end_date:
    :return:
    """
    import pyfi.function.DSfunctions as dsf
    # -----还原工业增加值绝对值-----
    IVA_idx = ip_idx(begin_date, end_date)
    IVA_Deseason = dsf.Deseason(IVA_idx.iloc[:, 0].dropna(how='all'), Jan_Feb=False)
    yoy_Deseason = (IVA_Deseason / IVA_Deseason.shift(12) - 1).dropna(how='all')
    return yoy_Deseason


def ip_idx(begin_date, end_date):
    """
    获得工业增加值去除春节影响的定基指数
    
    :param begin_date:
    :param end_date:
    :return:
    """
    from pyfi import WindHelper
    import pyfi.IVA_idx as ii
    df = WindHelper.edb(codes=['ip_yoy', 'ip_cyoy'], begin_date=begin_date, end_date=end_date)
    base = WindHelper.edb(codes=["M5567963"], begin_date=begin_date, end_date=end_date).iloc[:, 0]
    IVAyoy = df.loc[:, 'ip_yoy'] / 100.0
    IVAyoyc = df.loc[:, 'ip_cyoy'] / 100.0
    IVAyoyr = pd.DataFrame(1 + IVAyoy)
    IVAyoycr = pd.DataFrame(1 + IVAyoyc)
    Base = ii.mean_Jan_Feb(base)  # 平均基值的1、2月数据
    # -----还原工业增加值绝对值的定基指数-----
    IVA_idx = ii.get_idx(IVAyoyr, IVAyoycr, Base).dropna()
    return IVA_idx


def shuini(begin_date, end_date):
    """
    水泥指数
    :param begin_date:
    :param end_date:
    :return:
    """
    from pyfi import WindHelper
    df = WindHelper.edb(codes=["S5104572",
                               "S5104580",
                               "S5104577",
                               "S5104584",
                               "S5104588",
                               "S5104592"],
                        begin_date=begin_date,
                        end_date=end_date)
    return pd.DataFrame({"水泥指数": df.mean(axis=1)}, index=df.index)


def hbhdlm(begin_date, end_date):
    """
    环渤海动力煤(周度数据）
    :param begin_date:
    :param end_date:
    :return:
    """
    from pyfi import WindHelper
    df = WindHelper.edb(codes=["S5104572",
                               "S5104580",
                               "S5104577",
                               "S5104584",
                               "S5104588",
                               "S5104592"],
                        begin_date=begin_date,
                        end_date=end_date).ffill()
    return pd.DataFrame({"环渤海动力煤指数": df.mean(axis=1)}, index=df.index)


def get_inf_r_supseason(begin_date, end_date):
    from pyfi import WindHelper as w, get_supseason_ind
    name_list = ["cpi", "食品", "非食品", "核心", "不包括鲜菜和鲜果",
                 "消费品", "服务", "食品烟酒", "食品烟酒-粮食", "食品烟酒-食用油",
                 "食品烟酒-鲜菜", "食品烟酒-畜肉类", "食品烟酒-畜肉类-猪肉", "品烟酒-畜肉类-牛肉", "食品烟酒-畜肉类-羊肉",
                 "食品烟酒-水产品", "食品烟酒-蛋类", "食品烟酒-奶类", "食品烟酒-鲜果", "食品烟酒-烟草",
                 "食品烟酒-酒类", "衣着", "衣着-服装", "衣着-衣着加工服务费", "衣着-鞋类",
                 "居住", "居住-租赁房房租", "居住-水电燃费", "生活用品及服务", "生活用品及服务-家用电器",
                 "生活用品及服务-家庭服务", "交通和通信", "交通和通信-交通工具", "交通和通信-交通工具用燃料", "交通和通信-交通工具使用和维修",
                 "交通和通信-通信工具", "交通和通信-通信服务", "交通和通信-邮递服务", "教育文化和娱乐", "教育文化和娱乐-教育服务",
                 "教育文化和娱乐-旅游", "医疗保健", "医疗保健-中药", "医疗保健-西药", "医疗保健-医疗服务",
                 "其他用品和服务"
                 ]
    code_list = [
        "M0000705", "M0000706", "M0061581", "M0085934", "M0096670",
        "M0061583", "M0061585", "M0327907", "M0062906", "M0068106",
        "M0062910", "M0062907", "M0068107", "M0085938", "M0085942",
        "M0062909", "M0062908", "M0068169", "M0062911", "M0068108",
        "M0068109", "M0000708", "M0068110", "M0096676", "M0068170",
        "M0000713", "M0068121", "M0068122", "M0000709", "M0068111",
        "M0068112", "M0000711", "M0327927", "M0068116", "M0068117",
        "M0068118", "M0068171", "M0327917", "M0000712", "M0068172",
        "M0068119", "M0000710", "M0068113", "M0068114", "M0068115",
        "M0327908"
    ]
    df = w.edb(code_list, begin_date, end_date)  # .loc[:["M0068116"]]
    data = pd.DataFrame(index=df.index)
    data_cum = pd.DataFrame(index=df.index)
    for i in range(len(df.columns)):
        supseason_ind, supseason_cum_ind = get_supseason_ind(df.loc[:,code_list[i]])
        # 计算超季节性：
        item = pd.DataFrame({df.columns[i]: supseason_ind})
        data = data.join(item, how="outer")
        # 计算累计超季节性：
        item_cum = pd.DataFrame({df.columns[i]: supseason_cum_ind})
        data_cum = data_cum.join(item_cum, how="outer")
    # mapper = {key: value for key, value in zip(code_list, name_list)}
    data.columns = name_list
    data_cum.columns = name_list
    return data, data_cum




from pyfi.wind_helper import WindHelper
from pyfi.visual import double_lines, line_graph, multi_lines
from pyfi.mapper import *
from pyfi.template import *
from pyfi.common import *
from pyfi.processing import *
from pyfi.data_agent import *
from pyfi.IVA_idx import *
from pyfi.function.DSfunctions import *
from pyfi.function import *
from pyfi.backtest import monthly_backtest
from pyfi import logic



# encoding: utf-8
# python setup.py register sdist upload
from setuptools import setup, find_packages

VERSION = '0.1.20'

setup(name='pyfi_helper',
      version=VERSION,
      description='encapsulate the windPy API with pandas',
      url='https://github.com/wangluzhou/pyfi.git',
      author='wangluzhou',
      author_email='wangluzhou@aliyun.com',
      license='MIT',
      packages=find_packages(),
      zip_safe=True,
      install_requires=['pandas', 'matplotlib', 'scipy', 'matplotlib'])



# Author：Wang Luzhou
# Create: 2018-08-14
# 逻辑库
# 所有很宏观逻辑相关的库都在这里，蛤蛤
import numpy as np
import pandas as pd
# import talib

#  -----------事件原子逻辑算子------------------------

"""
零阶事件
"""


def zo1(data, args=[3]):
    """
    1个参数
    过去n个月的最大值
    :param data:
    :param args: args[0]: past n periods
    :return:
    """
    ind = pd.Series(index=data.index)
    past = args[0]
    for i in range(past - 1, len(data)):
        max = data[i - past + 1: i + 1].max()
        min = data[i - past + 1: i + 1].min()
        if data[i] == max:
            ind[i] = 1
        elif data[i] == min:
            ind[i] = -1
        else:
            ind[i] = 0
    return ind.dropna()


def zo2(data, args=(36, 0)):
    """

    :param data:
    :param args:
    :return:
    """
    past = args[0]
    threshold = args[1]
    if len(data) > 500:  # 说明是日度，直接在日度级别上进行标准化，防止信息丢失
        score = (data - data.rolling(window=past * 20).mean()) / data.rolling(window=past * 20).std()
        score = score.resample("M").mean().apply(lambda x: 0 if abs(x) < threshold else x)
    else:
        data = data.resample("M").mean()
        score = (data - data.rolling(window=past).mean()) / data.rolling(window=past).std()
        score = score.apply(lambda x: 0 if abs(x) < threshold else x)
    return score.dropna()


"""
一阶事件
"""


def fo1(data, args=[6]):
    """
    边际强度
    CPI： 2,2,2,2,2.5 产生最大的影响
    :param data:
    :param args:
    :return:
    """
    # 参数赋值
    # 计算分值
    from macroasset.first_logic import logic
    past = args[0]
    dif = pd.DataFrame(columns=range(1, past + 1), index=data.index)  # 存放斜率
    dif_b = pd.DataFrame(columns=range(1, past + 1), index=data.index)
    for i in dif.columns:
        dif.loc[:, i] = (data - data.shift(i)) / i
        dif_b.loc[:, i] = logic.one_signal(dif.loc[:, i])
    score_dif = dif_b.sum(axis=1)
    return score_dif


def fo2(data, args=[5]):
    """
    体现指标连续上涨的持续性
    :param data:
    :param args:
    :return:
    """
    past = args[0]
    dd = one_signal(data.diff().dropna())
    score = dd.rolling(window=past).sum()
    return score


def fo3(data, args=[1, 12]):
    """
    均线缺口本质是处于一阶和二阶之间的逻辑。在本框架内，我们将其视为1阶逻辑。
    2个参数
    data为series类型
    均线缺口,无截距
    args[0]:short period
    args[1]:long period
    args[2]: base line
    """
    data = data.ffill()
    ind = data.rolling(window=int(args[0])).mean() - data.rolling(window=int(args[1])).mean()
    return ind


def fo4(data, args=[1, 12, 1]):
    """
    3个参数
    :param data:
    :param args:0：短周期，1：长周期，2. 波动率截距
    :return:
    """
    data = data.ffill()
    short_mean = data.rolling(window=int(args[0])).mean()
    long_mean = data.rolling(window=int(args[1])).mean()
    vol_intercept = data.rolling(window=int(args[1])).std() * args[2]
    ind = short_mean - long_mean - vol_intercept
    return ind


def so1(data, args=[5]):
    """
    this logic contains first-order and second-order part, it is a mixed logic
    :param data:
    :param args: [0]: the longest length for past window; [1]:weight of the first order
    :return:
    """
    length = args[0]
    f_weight = 1
    d = []
    for i in range(1, length + 1):
        d.append((data - data.shift(i)) / i)  # reserve the nan to keep the same index
    # logic
    score = pd.Series(index=data.index)
    for j in range(length, len(data)):
        score[j] = f_weight * (1 if d[0][j] < 0 else (-1 if d[0][j] > 0 else 0))
        for i in range(len(d) - 1):
            score[j] = score[j] + (-1 if d[i][j] < d[i + 1][j] else (1 if d[i][j] > d[i + 1][j] else 0))
    score = score.dropna()
    return score


def so2(data, args=[5, 1, 1]):
    """
    this logic contains first-order and second-order part, it is a mixed logic
    :param data:
    :param args: [0]: the longest length for past window; [1]:weight of the first order
    :return:
    """
    length = args[0]
    f_weight = args[1]
    width = args[2]
    d = []
    for i in range(1, length + 1):
        d.append((data - data.shift(i * width)) / i)  # reserve the nan to keep the same index
    # logic
    score = pd.Series(index=data.index)
    for j in range(length, len(data)):
        score[j] = f_weight * (1 if d[0][j] < 0 else (-1 if d[0][j] > 0 else 0))
        for i in range(len(d) - 1):
            score[j] = score[j] + (-1 if d[i][j] < d[i + 1][j] else (1 if d[i][j] > d[i + 1][j] else 0))
    score = score.dropna()
    return score


def trend(data, args):
    """
    计算近期数据的拟合斜率
    :param data:
    :param args: 0：确定时间窗口的长短
    :return:
    """
    from sklearn import linear_model
    past = args[0]
    regr = linear_model.LinearRegression()
    s_data = pd.Series(index=data.index)
    for i in range(past, len(data)):
        regr.fit(np.array(list(range(past))).reshape(1, -1), data.values[i - past + 1: i + 1])
        s_data.iloc[i] = regr.coef_[0]
    return s_data.dropna()


def trend_count(data, args=[5]):
    """
    统计近几个月涨跌的净个数
    :param data:
    :param args:
    :return:
    """
    d = data.diff().dropna()
    count = args[0]
    ind = pd.Series(index=d.index)

    for i in range(count - 1, len(d)):
        ind[i] = d[i - count + 1: i + 1].apply(lambda x: 1 if x > 0 else (-1 if x < 0 else 0)).sum()
    return ind


def season(data, args):
    """
    季节性事件：当前的值和过去三年的均值进行对比，此类数据一般是环比月度数据
    :param data:
    :param args:
    :return:
    """
    ind = pd.Series()
    for i in range(1, 13):
        tmp = data.loc[data.index.month == i]
        tmp_mean = tmp.rolling(window=5).mean().shift(1)
        tmp_season = tmp / tmp_mean - 1 - args[0]
        ind.append(tmp_season)
    return ind.sort_index()


def mg3(data, args=[12, 1, 1]):
    """
    3个参数
    :param data:
    :param args:0：短周期，1：长周期，2. 波动率截距
    :return:
    """
    data = data.ffill()
    short_mean = data.rolling(window=int(args[0])).mean()
    long_mean = data.rolling(window=int(args[1])).mean()
    vol_intercept = data.rolling(window=int(args[1])).std() * args[2]
    ind = short_mean - long_mean - vol_intercept
    return ind


def mg2(data, args=[12, 1]):
    """
    2个参数
    data为series类型
    均线缺口,无截距
    args[0]:short period
    args[1]:long period
    args[2]: base line
    """
    data = data.ffill()
    ind = data.rolling(window=int(args[0])).mean() - data.rolling(window=int(args[1])).mean()
    return ind


# def cmo(data, args):
#     """
#     1个参数，趋势指标，返回Series
#     """
#     arg1 = args[0]
#     data = data.ffill()
#     ind = pd.Series(talib.CMO(data.values, timeperiod=arg1), index=data.index)
#     return ind

#
# def mom(data, args):
#     """
#     1个参数
#     momentum,多日的涨幅
#     :param data: pandas.Series
#     :param args: args[0]->timeperiod
#     :return: pandas.Series
#     """
#     data = data.ffill()
#     ind = pd.Series(talib.MOM(data.values, timeperiod=int(args[0])), index=data.index)
#     return ind


# def wmom(data, args):
#     """
#     3个参数
#     加权mom指标，多个不同时间端的涨幅的加权和
#     :param data:
#     :param args:
#     :return:
#     """
#     ind = (mom(data, int(args[0])).dropna()
#            + mom(data, int(args[1])).dropna()
#            + mom(data, int(args[2])).dropna()) / 3.0
#     return ind.dropna()


def slope(data, args):
    """
    区间头尾斜率
    :param data:
    :param args:0:近端斜率对应的区间， 1:远端斜率对应的区间
    :return:
    """
    past = args[0]
    slope = (data - data.shift(past - 1)) / past
    return slope


"""
2阶事件：反转事件
"""


def slopes(data, args=[5, 1, 1]):
    """
    this logic contains first-order and second-order part, it is a mixed logic
    :param data:
    :param args: [0]: the longest length for past window; [1]:weight of the first order
    :return:
    """
    length = args[0]
    f_weight = args[1]
    width = args[2]
    d = []
    for i in range(1, length + 1):
        d.append((data - data.shift(i * width)) / i)  # reserve the nan to keep the same index
    # logic
    score = pd.Series(index=data.index)
    for j in range(length, len(data)):
        score[j] = f_weight * (1 if d[0][j] < 0 else (-1 if d[0][j] > 0 else 0))
        for i in range(len(d) - 1):
            score[j] = score[j] + (-1 if d[i][j] < d[i + 1][j] else (1 if d[i][j] > d[i + 1][j] else 0))
    score = score.dropna()
    return score


def max_reversal(data, args):
    """
    高点前后两者斜率差
    :param data:
    :param args: 0 考察最近几个月的高点, 1. 截距
    :return:
    """
    # 找出高点
    ind = pd.Series(0.0, index=data.index)
    period = int(args[0])
    for i in range(period - 1, len(data)):
        tmp = data[i - period + 1:i]
        max_spot = tmp.max()
        idx = tmp.get_loc(max_spot)
        if idx != period - 2:
            continue
        else:
            ind.loc[ind.index == data.index[i]] = (tmp[-1] - tmp[idx]) / (tmp[0] - tmp[idx]) - 1 - args[1]
    return ind


def min_reversal(data, args):
    """
    触底反弹，低点前后两者斜率差
    :param data:
    :param args: 0 考察最近几个月的高点, 1. 截距
    :return:
    """
    # 找出高点
    ind = pd.Series(0.0, index=data.index)
    period = int(args[0])
    for i in range(period - 1, len(data)):
        tmp = data[i - period + 1:i]
        min_spot = tmp.min()
        idx = tmp.get_loc(min_spot)
        if idx != period - 2:
            continue
        else:
            ind.loc[ind.index == data.index[i]] = (tmp[-1] - tmp[idx]) / (tmp[0] - tmp[idx]) - 1 - args[1]
    return ind


def reversal(data, args):
    """
    连续N个周期上涨或下跌，最后一个交易日下跌或者上涨
    :param data:
    :param args:
    :return:
    """
    n = args[0]
    pct = data.pct_change()
    for i in range(n, len(pct)):
        if pct[i - n: i].sum() == n and pct[i] < 0:
            return -1.0
        elif pct[i - n: i].sum() == 0 and pct[i] > 0:
            return 1.0
        else:
            return 0.0


def slope_change(data, args):
    """
    两个区间的斜率变化差
    :param data:
    :param args:0:近端斜率对应的区间， 1:远端斜率对应的区间
    :return:
    """
    near_n = args[0]
    far_n = args[1]
    near_slope = (data - data.shift(near_n)) / near_n
    far_slope = (data - data.shift(far_n)) / far_n
    return near_slope - far_slope


def momentum(data, config, begin_date=None, end_date=None):
    """
    基于AQR Value and Momentum everywhere，过去指定时间的累积涨幅
    :param data:
    :param config: 0：past
    :param begin_date:
    :param end_date:
    :return:
    """
    past = config[0]
    mom = data / data.shift(past) - 1
    return mom.dropna()


# -------------简单N阶逻辑生成器----------------------------

def zero_order(data, begin_date=None, end_date=None):
    """
    :param data:
    :param begin_date:
    :param end_date:
    :return:
    """
    data = data.dropna()
    if begin_date is not None and end_date is not None:
        return data.loc[(data.index >= begin_date) & (data.index <= end_date)]
    else:
        return data


def first_order(data, begin_date=None, end_date=None):
    f_data = data.diff().dropna()
    if begin_date is not None and end_date is not None:
        return f_data.loc[(f_data.index >= begin_date) & (f_data.index <= end_date)]
    else:
        return f_data


def second_order(data, begin_date=None, end_date=None):
    s_data = data.diff().diff().dropna()
    if begin_date is not None and end_date is not None:
        return s_data.loc[(s_data.index >= begin_date) & (s_data.index <= end_date)]
    else:
        return s_data


# -------------逻辑长度算子----------------------------
def impulse(data, event_date, begin_date=None, end_date=None, args=[1]):
    """
    冲激
    :param data:
    :param event_date:
    :param begin_date:
    :param end_date:
    :param args: [height]
    :return:
    """
    height = args[0]
    data.loc[event_date] += height
    return slice(data, begin_date, end_date)


def step(data, event_date, begin_date=None, end_date=None, args=[1]):
    """
    阶跃
    :param data:
    :param event_date:
    :param begin_date:
    :param end_date:
    :param args: [height]
    :return:
    """
    height = args[0]
    data.loc[data.index >= event_date] += height
    return slice(data, begin_date, end_date)


def decay(data, event_date, positive=True, begin_date=None, end_date=None, args=[5, 2, 1]):
    # to add a new signal shape to the original signal series "data"
    """
    衰减
    :param data:
    :param event_date:
    :param positive
    :param begin_date:
    :param end_date:
    :param args: [height, hold, rate]
    :return:
    """
    height = args[0]
    hold = args[1]
    rate = args[2]
    length = int(np.ceil(height / rate))
    try:
        idx = data.index.get_loc(event_date)  # get the index of date of the event occurence
    except Exception as e:
        print("The event date does not exist in the index of the data!!!")
        raise e
    # construct the shape of the signal
    shape = [height] * (1 + hold)
    shape += [max(height - rate * (i + 1), 0) for i in range(length)]
    shape = np.array(shape) * (1 if positive else -1)
    # slice the shape
    last_idx = min(idx + hold + length, len(data) - 1)
    shape = shape[0: last_idx - idx + 1]
    data.iloc[idx: last_idx + 1] = shape
    return slice(data, begin_date, end_date)


def half_life(data, T=30):
    """
    半衰期
    :param data:
    :param T:
    :return:
    """

    pass


#  ------------信号处理逻辑算子-------------------------

def cumsum(score, up=5, low=-5):
    """
    将信号逐步叠加
    :return:
    """
    score = score.fillna(0)
    cum_score = pd.Series(index=score.index)
    cum_score[0] = score[0]
    for i in range(1, len(cum_score)):
        cum_score[i] = cap(cum_score[i - 1] + score[i], low=low, up=up)
    return cum_score.dropna()


def cap(score, up=5, low=-5):
    """
    限定信号的最大值和最小值，防止信号超过阈值范围
    :param score: float or series
    :param up:
    :param low:
    :return:
    """
    if hasattr(score, "__iter__"):
        return score.apply(lambda x: min(max(x, low), up))
    return min(max(score, low), up)


def value_score(value, up, low):
    """
    针对估值指标的打分，分值具有历史惯性
    :param value: Series
    :param up:
    :param low:
    :return:
    """
    score = pd.Series(index=value.index)
    for i in range(len(score)):
        if i == 0:
            if value[i] > up:
                score[i] = 1
            elif value[i] < low:
                score[i] = -1
            else:
                score[i] = 0
        else:
            if value[i] > up:
                score[i] = 1
            elif value[i] <= low:
                score[i] = -1
            elif score[i - 1] == 1:
                if value[i] <= 0:
                    score[i] = 0  # 维持基准
                elif value[i] > 0:
                    score[i] = score[i - 1]
            elif score[i - 1] == -1:
                if value[i] >= 0:
                    score[i] = 0
                elif value[i] < 0:
                    score[i] = score[i - 1]
            else:
                score[i] = 0
    return score


def to_score(low, up, max=5, min=-5, center=0):
    pass


# -----------一般操作符--------------------------------
def slice(data, begin_date=None, end_date=None):
    if begin_date is not None:
        data = data.loc[(data.index >= begin_date)]
    if end_date is not None:
        data = data.loc[(data.index <= end_date)]
    return data


def shift(data, s=1):
    return data.shift(s)


# ------------逻辑工厂-----------------------------


def index_shift(data, shift=1, freq="M"):
    data.index = data.index.shift(shift, freq=freq)
    return data


def one_signal(data, positive=True, precious=1e-10):
    """
    将指标数据转换为-1,0，+1三维信号模式， 注意已包含去na功能
    :param precious:
    :param positive:
    :param data: pandas.Series
    :return: pandas.Series
    """
    d = 1 if positive else -1
    return data.apply(lambda x: 1 if x > precious else (-1 if x < -precious else 0)).dropna() * d


def one_signal_for_df(df):
    """
    二元信号处理
    :param df:
    :return:
    """
    df = df.copy()
    for i in range(len(df.columns)):
        df.iloc[:, i] = one_signal(df.iloc[:, i])
    return df


def signal_trans(score, frm=5, to=2):
    """
    :param score:
    :param frm: 过去的分制
    :param to: 目标分制
    :return:
    """
    return score.apply(lambda x: round(x * 1.0 * to / frm))
